# -*- encoding: utf-8 -*-
"""
keri.kli.commands module

"""
import argparse
import getpass
import os
import sys

from hio import help
from hio.base import doing

import keri.app.oobiing
from keri import kering
from keri.app import habbing, configing, oobiing
from keri.app.keeping import Algos
from keri.db import basing
from keri.kering import ConfigurationError
from keri.vdr import credentialing

logger = help.ogler.getLogger()


def handler(args):
    """
    Launch KERI database initialization

    Args:
        args(Namespace): arguments object from command line
    """
    init = InitDoer(args)
    return [init]


parser = argparse.ArgumentParser(description='Create a database and keystore')
parser.set_defaults(handler=handler,
                    transferable=True)

# Parameters for basic structure of database
parser.add_argument('--name', '-n', help='keystore name and file location of KERI keystore', required=True)
parser.add_argument('--base', '-b', help='additional optional prefix to file location of KERI keystore',
                    required=False, default="")
parser.add_argument('--temp', '-t', help='create a temporary keystore, used for testing', default=False)
parser.add_argument('--salt', '-s', help='qualified base64 salt for creating key pairs', required=False)
parser.add_argument("--config-dir", "-c", dest="configDir", help="directory override for configuration data")
parser.add_argument('--config-file',
                    dest="configFile",
                    action='store',
                    default=None,
                    help="configuration filename override")

# Parameters for Manager creation
# passcode => bran
parser.add_argument('--passcode', '-p', help='21 character encryption passcode for keystore (is not saved)',
                    dest="bran", default=None)
parser.add_argument('--nopasscode', help='create an unencrypted keystore', action='store_true')
parser.add_argument('--aeid', '-a', help='qualified base64 of non-transferable identifier prefix for  authentication '
                                         'and encryption of secrets in keystore', default=None)
parser.add_argument('--seed', '-e', help='qualified base64 private-signing key (seed) for the aeid from which the '
                                         'private decryption key may be derived', default=None)


class InitDoer(doing.DoDoer):

    def __init__(self, args):
        self.args = args
        super(InitDoer, self).__init__(doers=[doing.doify(self.initialize)])

    def initialize(self, tymth, tock=0.0):

        # enter context
        self.wind(tymth)
        self.tock = tock
        _ = (yield self.tock)

        args = self.args
        name = args.name
        if name is None or name == "":
            raise ConfigurationError("Name is required and can not be empty")

        base = args.base
        temp = args.temp
        bran = args.bran
        configFile = args.configFile
        configDir = args.configDir

        if not args.nopasscode and not bran:
            print("Creating encrypted keystore, please enter your 21 character passcode:")
            while True:
                bran = getpass.getpass("Passcode: ")
                retry = getpass.getpass("Re-enter passcode: ")

                if bran != retry:
                    print("Passcodes do not match, try again.")
                else:
                    break

        kwa = dict()
        kwa["salt"] = args.salt
        kwa["bran"] = bran
        kwa["aeid"] = args.aeid
        kwa["seed"] = args.seed
        if args.salt is None:
            kwa["algo"] = Algos.randy

        cf = None
        if configFile is not None:
            cf = configing.Configer(name=configFile,
                                    base="",
                                    headDirPath=configDir,
                                    temp=False,
                                    reopen=True,
                                    clear=False)

        hby = habbing.Habery(name=name, base=base, temp=temp, cf=cf, **kwa)
        rgy = credentialing.Regery(hby=hby, name=name, base=base, temp=temp)

        print("KERI Keystore created at:", hby.ks.path)
        print("KERI Database created at:", hby.db.path)
        print("KERI Credential Store created at:", rgy.reger.path)
        if hby.mgr.aeid:
            print("\taeid:", hby.mgr.aeid)

        oc = hby.db.oobis.cntAll()
        if oc:
            print(f"\nLoading {oc} OOBIs...")

            obi = keri.app.oobiing.Oobiery(hby=hby)
            self.extend(obi.doers)

            while oc > hby.db.roobi.cntAll():
                yield 0.25

            for (oobi,), obr in hby.db.roobi.getItemIter():
                if obr.state in (oobiing.Result.resolved,):
                    print(oobi, "succeeded")
                if obr in (oobiing.Result.failed,):
                    print(oobi, "failed")

            self.remove(obi.doers)

        wc = [oobi for (oobi,), _ in hby.db.woobi.getItemIter()]
        if len(wc) > 0:
            print(f"\nAuthenticating Well-Knowns...")
            authn = oobiing.Authenticator(hby=hby)
            self.extend(authn.doers)

            while True:
                cap = []
                for (_,), wk in hby.db.wkas.getItemIter(keys=b''):
                    cap.append(wk.url)

                if set(wc) & set(cap) == set(wc):
                    break

                yield 0.5

            self.remove(authn.doers)

        hby.close()



# -*- encoding: utf-8 -*-
"""
keri.kli.commands module

"""
import argparse

from hio.base import doing

from keri import kering
from keri.app.cli.common import existing
from keri.core import indexing, coring, MtrDex

parser = argparse.ArgumentParser(description='Decrypt arbitrary data for AIDs with Ed25519 public keys only')
parser.set_defaults(handler=lambda args: handler(args))
parser.add_argument('--name', '-n', help='keystore name and file location of KERI keystore', required=True)
parser.add_argument('--base', '-b', help='additional optional prefix to file location of KERI keystore',
                    required=False, default="")
parser.add_argument('--alias', '-a', help='human readable alias for the new identifier prefix', required=True)
parser.add_argument('--passcode', '-p', help='22 character encryption passcode for keystore (is not saved)',
                    dest="bran", default=None)  # passcode => bran
parser.add_argument('--data', '-d', help='Encrypted data or file (starts with "@")', required=True)


def handler(args):
    """
    Verify signatures on arbitrary data

    Args:
        args(Namespace): arguments object from command line
    """
    kwa = dict(args=args)
    return [doing.doify(decrypt, **kwa)]


def decrypt(tymth, tock=0.0, **opts):
    """ Command line status handler

    """
    _ = (yield tock)
    args = opts["args"]

    name = args.name
    alias = args.alias
    base = args.base
    bran = args.bran

    try:
        with existing.existingHab(name=name, alias=alias, base=base, bran=bran) as (_, hab):

            data = args.data
            if data.startswith("@"):
                f = open(data[1:], "r")
                data = f.read()
            else:
                data = data

            m = coring.Matter(qb64=data)  # should refactor this to use Cipher
            d = coring.Matter(qb64=hab.decrypt(ser=m.raw))
            print(d.raw)

    except kering.ConfigurationError:
        print(f"prefix for {name} does not exist, incept must be run first", )
    except FileNotFoundError:
        print("unable to open file", args.text[1:])



# -*- encoding: utf-8 -*-
"""
keri.kli.commands module

"""
import argparse

from hio.base import doing

from keri.app import habbing, forwarding, connecting
from keri.app.cli.common import existing
from keri.app.habbing import GroupHab
from keri.peer import exchanging

parser = argparse.ArgumentParser(description='Respond to a list of challenge words by signing and sending an EXN '
                                             'response')
parser.set_defaults(handler=lambda args: respond(args))
parser.add_argument('--name', '-n', help='keystore name and file location of KERI keystore', required=True)
parser.add_argument('--base', '-b', help='additional optional prefix to file location of KERI keystore',
                    required=False, default="")
parser.add_argument('--alias', '-a', help='human readable alias for the new identifier prefix', default=None)
parser.add_argument('--passcode', '-p', help='21 character encryption passcode for keystore (is not saved)',
                    dest="bran", default=None)  # passcode => bran
parser.add_argument('--words', '-d', help='JSON formatted array of words to sign, \'@\' allowed to load from a file',
                    action="store", required=True)
parser.add_argument('--recipient', '-r', help='Contact alias of the AID to send the signed words to',
                    action="store", required=True)


def respond(args):
    """
    Performs a rotation of the identifier of the environment represented by the provided name parameter

        args (parseargs):  Command line argument

    """
    name = args.name
    alias = args.alias
    base = args.base
    bran = args.bran
    recp = args.recipient

    if args.words.startswith("@"):
        f = open(args.data[1:], "r")
        words = f.read()
    else:
        words = args.words

    words = words.split(" ")
    if not isinstance(words, list):
        raise ValueError("words must be an array of words")

    ixnDoer = RespondDoer(name=name, base=base, alias=alias, bran=bran, words=words, recp=recp)

    return [ixnDoer]


class RespondDoer(doing.DoDoer):
    """
    DoDoer that launches Doers needed to create an interaction event and publication of the event
    to all appropriate witnesses
    """

    def __init__(self, name, base, bran, alias, words: list, recp: str):
        """
        Returns DoDoer with all registered Doers needed to perform interaction event.

        Parameters:
            name is human readable str of identifier
            proto is tcp or http method for communicating with Witness
            data is list of dicts of committed data such as seals
       """

        self.alias = alias
        self.words = words
        self.recp = recp

        self.hby = existing.setupHby(name=name, base=base, bran=bran)
        self.postman = forwarding.Poster(hby=self.hby)
        self.hbyDoer = habbing.HaberyDoer(habery=self.hby)  # setup doer
        self.org = connecting.Organizer(hby=self.hby)
        doers = [self.hbyDoer, self.postman, doing.doify(self.respondDo)]

        super(RespondDoer, self).__init__(doers=doers)

    def respondDo(self, tymth, tock=0.0, **opts):
        """
        Returns:  doifiable Doist compatible generator method
        Usage:
            add result of doify on this method to doers list
        """
        self.wind(tymth)
        self.tock = tock
        _ = (yield self.tock)

        if self.alias is None:
            self.alias = existing.aliasInput(self.hby)

        hab = self.hby.habByName(name=self.alias)
        if hab is None:
            raise ValueError(f"invalid alias {self.alias}")

        recp = self.org.find("alias", self.recp)
        if len(recp) != 1:
            raise ValueError(f"invalid recipient {self.recp}")

        recp = recp[0]['id']

        payload = dict(i=hab.pre, words=self.words)
        exn, _ = exchanging.exchange(route="/challenge/response", payload=payload, sender=hab.pre)
        ims = hab.endorse(serder=exn, last=False, pipelined=False)
        del ims[:exn.size]

        senderHab = hab.mhab if isinstance(hab, GroupHab) else hab
        self.postman.send(src=senderHab.pre, dest=recp, topic="challenge", serder=exn, attachment=ims)
        while not self.postman.cues:
            yield self.tock

        toRemove = [self.hbyDoer, self.postman]
        self.remove(toRemove)

        return



# -*- encoding: utf-8 -*-
"""
keri.kli.commands module

"""
import argparse
from dataclasses import dataclass

from hio import help
from hio.base import doing

from keri.app import habbing, agenting, indirecting, configing, delegating, forwarding
from keri.app.cli.common import existing, incepting, config
from keri.core import coring

logger = help.ogler.getLogger()

parser = argparse.ArgumentParser(description='Initialize a prefix')
parser.set_defaults(handler=lambda args: handler(args))
parser.add_argument('--name', '-n', help='keystore name and file location of KERI keystore', required=True)
parser.add_argument('--base', '-b', help='additional optional prefix to file location of KERI keystore',
                    required=False, default="")
parser.add_argument('--alias', '-a', help='human readable alias for the new identifier prefix', required=True)
parser.add_argument("--config", "-c", help="directory override for configuration data")
parser.add_argument("--receipt-endpoint", help="Attempt to connect to witness receipt endpoint for witness receipts.",
                    dest="endpoint", action='store_true')
parser.add_argument("--proxy", help="alias for delegation communication proxy", default="")

parser.add_argument('--file', '-f', help='Filename to use to create the identifier', default="", required=False)

# Authentication for keystore
parser.add_argument('--passcode', '-p', help='21 character encryption passcode for keystore (is not saved)',
                    dest="bran", default=None)  # passcode => bran
parser.add_argument('--aeid', help='qualified base64 of non-transferable identifier prefix for  authentication '
                                   'and encryption of secrets in keystore', default=None)
incepting.addInceptingArgs(parser)


@dataclass
class InceptOptions:
    """ Options loaded from file parameter.

    """
    transferable: bool | None
    wits: list | None
    icount: int | None
    isith: int | str | list | None
    ncount: int | None
    nsith: int | str | list | None = '0'
    toad: int = 0
    delpre: str = None
    estOnly: bool = False
    data: list = None


def handler(args):
    """
    Create KERI identifier prefix in specified key store with alias

    Args:
        args(Namespace): arguments object from command line
    """

    name = args.name
    base = args.base
    bran = args.bran
    alias = args.alias
    config_dir = args.config
    endpoint = args.endpoint
    proxy = args.proxy

    kwa = mergeArgsWithFile(args).__dict__

    icpDoer = InceptDoer(name=name, base=base, alias=alias, bran=bran, endpoint=endpoint, proxy=proxy,
                         cnfg=config_dir, **kwa)

    doers = [icpDoer]
    return doers


def emptyOptions():
    """
    Initializes an empty inception options to be used only when required values are passed in at the command line
    """
    return InceptOptions(
        transferable=None, wits=None, icount=None, isith=None, ncount=None, nsith=None
    )


def mergeArgsWithFile(args):
    """
    Merge options specified with command line arguments with any specified config file
    with command line arguments taking precedence.
    """
    required_args = ['transferable', 'wits', 'icount', 'isith', 'ncount', 'nsith', 'toad']
    if args.file is None or args.file == '':
        config.checkRequiredArgs(args, required_args)

    incept_opts = config.loadFileOptions(args.file, InceptOptions) if args.file != '' else emptyOptions()

    incept_opts.transferable = True if args.transferable else incept_opts.transferable
    if len(args.wits) > 0:
        incept_opts.wits = args.wits
    if args.icount is not None:
        incept_opts.icount = int(args.icount)
    if args.toad is not None:
        incept_opts.toad = int(args.toad)
    if args.icount is not None:
        incept_opts.icount = int(args.icount)
    if args.isith is not None:
        incept_opts.isith = args.isith
    if args.ncount is not None:
        incept_opts.ncount = int(args.ncount)
    if args.nsith is not None:
        incept_opts.nsith = args.nsith
    if args.est_only is not None:
        incept_opts.estOnly = args.est_only
    if args.data is not None:
        incept_opts.data = config.parseData(args.data)
    if args.delpre is not None:
        incept_opts.delpre = args.delpre

    return incept_opts


class InceptDoer(doing.DoDoer):
    """ DoDoer for creating a new identifier prefix and Hab with an alias.
    """

    def __init__(self, name, base, alias, bran, endpoint, proxy=None, cnfg=None, **kwa):

        cf = None
        if config is not None:
            cf = configing.Configer(name=name,
                                    base="",
                                    headDirPath=cnfg,
                                    temp=False,
                                    reopen=True,
                                    clear=False)
        self.endpoint = endpoint
        self.hby = existing.setupHby(name=name, base=base, bran=bran, cf=cf)
        self.proxy = self.hby.habByName(proxy) if proxy is not None else None
        self.hbyDoer = habbing.HaberyDoer(habery=self.hby)  # setup doer
        self.swain = delegating.Anchorer(hby=self.hby, proxy=self.proxy)
        self.postman = forwarding.Poster(hby=self.hby)
        self.mbx = indirecting.MailboxDirector(hby=self.hby, topics=['/receipt', "/replay", "/reply"])
        doers = [self.hbyDoer, self.postman, self.mbx, self.swain, doing.doify(self.inceptDo)]

        self.inits = kwa
        self.alias = alias
        super(InceptDoer, self).__init__(doers=doers)

    def inceptDo(self, tymth, tock=0.0):
        """
        Parameters:
            tymth (function): injected function wrapper closure returned by .tymen() of
                Tymist instance. Calling tymth() returns associated Tymist .tyme.
            tock (float): injected initial tock value

        Returns:  doifiable Doist compatible generator method
        """
        # enter context
        self.wind(tymth)
        self.tock = tock
        _ = (yield self.tock)

        hab = self.hby.makeHab(name=self.alias, **self.inits)
        witDoer = agenting.WitnessReceiptor(hby=self.hby)
        receiptor = agenting.Receiptor(hby=self.hby)
        self.extend([witDoer, receiptor])

        if hab.kever.delpre:
            self.swain.delegation(pre=hab.pre, sn=0)
            print("Waiting for delegation approval...")
            while not self.swain.complete(hab.kever.prefixer, coring.Seqner(sn=hab.kever.sn)):
                yield self.tock

        elif hab.kever.wits:
            print("Waiting for witness receipts...")
            if self.endpoint:
                yield from receiptor.receipt(hab.pre, sn=0)
            else:
                witDoer.msgs.append(dict(pre=hab.pre))
                while not witDoer.cues:
                    _ = yield self.tock

        if hab.kever.delpre:
            if self.proxy is not None:
                sender = self.proxy
            else:
                sender = hab
            yield from self.postman.sendEventToDelegator(hab=hab, sender=sender, fn=hab.kever.sn)

        print(f'Prefix  {hab.pre}')
        for idx, verfer in enumerate(hab.kever.verfers):
            print(f'\tPublic key {idx + 1}:  {verfer.qb64}')
        print()

        toRemove = [self.hbyDoer, witDoer, self.mbx, self.swain, self.postman, receiptor]
        self.remove(toRemove)

        return



# -*- encoding: utf-8 -*-
"""
keri.kli.commands module

"""
import argparse

from hio import help
from hio.base import doing

from keri.app.cli.common import existing

logger = help.ogler.getLogger()


def handler(args):
    """
    Launch KERI database initialization

    Args:
        args(Namespace): arguments object from command line
    """
    clean = CleanDoer(args)
    return [clean]


parser = argparse.ArgumentParser(description='Cleans and migrates a database and keystore')
parser.set_defaults(handler=handler,
                    transferable=True)

# Parameters for basic structure of database
parser.add_argument('--name', '-n', help='keystore name and file location of KERI keystore', required=True)
parser.add_argument('--base', '-b', help='additional optional prefix to file location of KERI keystore',
                    required=False, default="")
parser.add_argument('--temp', '-t', help='create a temporary keystore, used for testing', default=False)
parser.add_argument('--migration', '-m', help='migration name', required=True)


# Parameters for Manager creation
# passcode => bran
parser.add_argument('--passcode', '-p', help='21 character encryption passcode for keystore (is not saved)',
                    dest="bran", default=None)


class CleanDoer(doing.Doer):

    def __init__(self, args):
        self.args = args
        super(CleanDoer, self).__init__()

    def recur(self, tyme):
        hby = existing.setupHby(name=self.args.name, base=self.args.base,
                                bran=self.args.bran, temp=self.args.temp)

        [(name, dater)] = hby.db.complete(name=self.args.migration)
        date = dater.datetime.strftime("%Y-%m-%d %H:%M") if dater is not None else "Not Run"

        print(f"{self.args.migration} -> {date}")

        return True



# -*- encoding: utf-8 -*-
"""
KERI
keri.db.subing module

Provide variety of mixin classes for LMDB sub-dbs with various behaviors.

Principally:

Suber class provides put, pin, get, rem and getItemIter method for managing
a serialized value in a sub db with an iterable set of keys defining the key space

CesrSuber class extends Suber for values that are serializations of CESR serializable
object instances. Ducktyped subclasses of Matter, Indexer, and Counter or the like.

IoSetSuber class extends Suber to allow a set of values to be stored in insertion
order at each effective key. Only one copy of a unique value is allowed in the
set at a given effective key. The effective key suffixes an ordinal to the key
space to track insertion ordering. IoSetSuber adds additional methods to manage
IoSets of values.

CatCesrSuber adds the ability to store multiple concatenated serializations at
a value

CatCesrIoSetSuber combines the capabilities

Other special classer for special values

SerderSuber stores Serialized Serder Instances of in JSON, CBOR, or MGPK

Also for Secrets private keys
SignerSuber
CryptSignerSuber

Also for using the dupsort==true mechanism is
DupSuber
CesrDupSuber

Class Architecture

Suber is simple lexographic database with only one value per key
OnSuber is simple lexographic database where trailing part of key is serialized
    ordinal number so that the ordering within each key prefix is monotonically
    increasing numeric

The term 'set' of values means that no value may appear more than once in the set.
Sets support idempotent adds and puts to db. This means one can add or put the same
(key, val) pair multiple times and not change the db.

DupSuber provides set of lexicographic ordered values at each key. Each value has
    a limited size (key + value <= 511 byes). The set is performant. Good for indices.

IoDupSuber provides set of insertion ordered values at each key. Each value has
    a limited size (key + value <= 511 byes). The set is less perfromant than DupSuber
    but more performant than IoSetSuber. Good for insertion ordered indices

IoSetSuber proves set of insertion ordered values at each key. Value size is not limited
    Good for any insertion ordered set where size may be too large for IoDupSuber

OnIoDupSuber provides set of insertion ordered values where the where trailing
    part of key is serialized ordinal number so that the ordering within each
    key prefix is monotonically increasing numeric.

Each of these base types for managing the key space may be mixed with other
Classes that provide different types of values these include.

Cesr
CatCesr
Serder
etc.


"""
from typing import Type, Union
from collections.abc import Iterable, Iterator

from .. import help
from ..help.helping import nonStringIterable, Reb64
from .. import core
from ..core import coring, scheming, serdering
from . import dbing

logger = help.ogler.getLogger()


class SuberBase():
    """
    Base class for Sub DBs of LMDBer
    Provides common methods for subclasses
    Do not instantiate but use a subclass

    Attributes:
        db (dbing.LMDBer): base LMDB db
        sdb (lmdb._Database): instance of lmdb named sub db for this Suber
        sep (str): separator for combining keys tuple of strs into key bytes
        verify (bool): True means reverify when ._des from db when applicable
                       False means do not reverify. Default False
    """
    Sep = '.'  # separator for combining key iterables

    def __init__(self, db: dbing.LMDBer, *,
                       subkey: str='docs.',
                       dupsort: bool=False,
                       sep: str=None,
                       verify: bool=False,
                       **kwa):
        """
        Parameters:
            db (dbing.LMDBer): base db
            subkey (str):  LMDB sub database key
            dupsort (bool): True means enable duplicates at each key
                               False (default) means do not enable duplicates at
                               each key
            sep (str): separator to convert keys iterator to key bytes for db key
                       default is self.Sep == '.'
            verify (bool): True means reverify when ._des from db when applicable
                           False means do not reverify. Default False
        """
        super(SuberBase, self).__init__()  # for multi inheritance
        self.db = db
        self.sdb = self.db.env.open_db(key=subkey.encode("utf-8"), dupsort=dupsort)
        self.sep = sep if sep is not None else self.Sep
        self.verify = True if verify else False



    def _tokey(self, keys: str|bytes|memoryview|Iterable[str|bytes|memoryview],
                topive: bool=False):
        """
        Converts keys to key bytes with proper separators and returns key bytes.
        If keys is already str or bytes or memoryview then returns key bytes.
        Else If keys is iterable (non-str) of strs or bytes then joins with
        separator converts to key bytes and returns. When keys is iterable and
        topive is True then enables partial key from top branch of key space given
        by partial keys by appending separator to end of partial key

        Returns:
           key (bytes): each element of keys is joined by .sep. If topive then
                        last char of key is .sep

        Parameters:
           keys (str | bytes | memoryview | Iterable[str | bytes]): db key or
                        Iterable of (str | bytes) to form key.
           topive (bool): True means treat as partial key tuple from top branch of
                       key space given by partial keys. Resultant key ends in .sep
                       character.
                       False means treat as full branch in key space. Resultant key
                       does not end in .sep character.
                       When last item in keys is empty str then will treat as
                       partial ending in sep regardless of topive value

        """
        if hasattr(keys, "encode"):  # str
            return keys.encode("utf-8")
        if isinstance(keys, memoryview):  # memoryview of bytes
            return bytes(keys)  # return bytes
        elif hasattr(keys, "decode"): # bytes
            return keys
        if topive and keys[-1]:  # topive and keys is not already partial
            keys = tuple(keys) + ('',)  # cat empty str so join adds trailing sep
        return (self.sep.join(key.decode() if hasattr(key, "decode") else key
                              for key in keys).encode("utf-8"))


    def _tokeys(self, key: str | bytes | memoryview):
        """
        Converts key bytes to keys tuple of strs by decoding and then splitting
        at separator .sep.

        Returns:
           keys (tuple[str]): makes tuple by splitting key at sep

        Parameters:
           key (str | bytes | memoryview): db key.

        """
        if isinstance(key, memoryview):  # memoryview of bytes
            key = bytes(key)
        if hasattr(key, "decode"):  # bytes
            key = key.decode("utf-8")  # convert to str
        return tuple(key.split(self.sep))


    def _ser(self, val: str | bytes | memoryview):
        """
        Serialize value to bytes to store in db
        Parameters:
            val (str | bytes | memoryview): encodable as bytes
        """
        if isinstance(val, memoryview):  # memoryview is always bytes
            val = bytes(val)  # return bytes
        return (val.encode("utf-8") if hasattr(val, "encode") else val)


    def _des(self, val: bytes | memoryview):
        """
        Deserialize val to str
        Parameters:
            val (bytes | memoryview): decodable as str
        """
        if isinstance(val, memoryview):  # memoryview is always bytes
            val = bytes(val)  # convert to bytes
        return (val.decode("utf-8") if hasattr(val, "decode") else val)


    def trim(self, keys: str|bytes|memoryview|Iterable=b"", *, topive=False):
        """
        Removes all entries whose keys startswith keys. Enables removal of whole
        branches of db key space. To ensure that proper separation of a branch
        include empty string as last key in keys. For example ("a","") deletes
        'a.1'and 'a.2' but not 'ab'

        Parameters:
            keys (Iteratabke[str | bytes | memoryview]): of key parts that may be
                a truncation of a full keys tuple in  in order to address all the
                items from multiple branches of the key space.
                If keys is empty then trims all items in database.
                Either append "" to end of keys Iterable to ensure get properly
                separated top branch key or use top=True.

            topive (bool): True means treat as partial key tuple from top branch of
                       key space given by partial keys. Resultant key ends in .sep
                       character.
                       False means treat as full branch in key space. Resultant key
                       does not end in .sep character.
                       When last item in keys is empty str then will treat as
                       partial ending in sep regardless of top value


        Returns:
           result (bool): True if val at key exists so delete successful. False otherwise
        """
        return(self.db.delTopVal(db=self.sdb, top=self._tokey(keys, topive=topive)))


    def getFullItemIter(self, keys: str|bytes|memoryview|Iterable[str|bytes]="",
                       *, topive=False):
        """Iterator over items in .db that returns full items with subclass
        specific special hidden parts shown for debugging or testing.

        Returns:
            items (Iterator[tuple[key,val]]): (key, val) tuples of each item
            over the all the items in subdb whose key startswith key made from
            keys. Keys may be keyspace prefix to return branches of key space.
            When keys is empty then returns all items in subdb.
            This is meant to return full parts of items in both keyspace and
            valuespace which may be useful in debugging or testing.

        Parameters:
            keys (str|bytes|memoryview|Iteratable[str | bytes | memoryview]):
                of key parts that may be
                a truncation of a full keys tuple in  in order to address all the
                items from multiple branches of the key space.
                If keys is empty then gets all items in database.
                Either append "" to end of keys Iterable to ensure get properly
                separated top branch key or use top=True.
                In Python str.startswith('') always returns True so if branch
                key is empty string it matches all keys in db with startswith.


            topive (bool): True means treat as partial key tuple from top branch of
                       key space given by partial keys. Resultant key ends in .sep
                       character.
                       False means treat as full branch in key space. Resultant key
                       does not end in .sep character.
                       When last item in keys is empty str then will treat as
                       partial ending in sep regardless of top value

        """
        for key, val in self.db.getTopItemIter(db=self.sdb,
                                               top=self._tokey(keys, topive=topive)):
            yield (self._tokeys(key), self._des(val))


    def getItemIter(self, keys: str|bytes|memoryview|Iterable="",
                       *, topive=False):
        """Iterator over items in .db subclasses that do special hidden transforms
        on either the keyspace or valuespace should override this method to hide
        hidden parts from the returned items. For example, adding either
        a hidden key space suffix or hidden val space proem to ensure insertion
        order. Use getFullItemIter instead to return full items with hidden parts
        shown for debugging or testing.

        Returns:
            items (Iterator[tuple[key,val]]): (key, val) tuples of each item
            over the all the items in subdb whose key startswith key made from
            keys. Keys may be keyspace prefix to return branches of key space.
            When keys is empty then returns all items in subdb



        Parameters:
            keys (str|bytes|memoryview|Iterable[str|bytes|memoryview]): of key
                parts that may be
                a truncation of a full keys tuple in  in order to address all the
                items from multiple branches of the key space.
                If keys is empty then gets all items in database.
                Either append "" to end of keys Iterable to ensure get properly
                separated top branch key or use top=True.
                In Python str.startswith('') always returns True so if branch
                key is empty string it matches all keys in db with startswith.


            topive (bool): True means treat as partial key tuple from top branch of
                key space given by partial keys. Resultant key ends in .sep
                character.
                False means treat as full branch in key space. Resultant key
                does not end in .sep character.
                When last item in keys is empty str then will treat as
                partial ending in sep regardless of top value

        """
        for key, val in self.db.getTopItemIter(db=self.sdb,
                                               top=self._tokey(keys, topive=topive)):
            yield (self._tokeys(key), self._des(val))


class Suber(SuberBase):
    """
    Subclass of SuberBase with no LMDB duplicates (i.e. multiple values at same key).
    """

    def __init__(self, db: dbing.LMDBer, *,
                       subkey: str = 'docs.',
                       dupsort: bool=False, **kwa):
        """
        Inherited Parameters:
            db (dbing.LMDBer): base db
            subkey (str):  LMDB sub database key
            dupsort (bool): True means enable duplicates at each key
                               False (default) means do not enable duplicates at
                               each key. Set to False
            sep (str): separator to convert keys iterator to key bytes for db key
                       default is self.Sep == '.'
            verify (bool): True means reverify when ._des from db when applicable
                           False means do not reverify. Default False

        Parameters:
            db (dbing.LMDBer): base db
            subkey (str):  LMDB sub database key
        """
        super(Suber, self).__init__(db=db, subkey=subkey, dupsort=False, **kwa)


    def put(self, keys: Union[str, Iterable], val: Union[bytes, str, any]):
        """
        Puts val at key made from keys. Does not overwrite

        Parameters:
            keys (tuple): of key strs to be combined in order to form key
            val (bytes): value

        Returns:
            result (bool): True If successful, False otherwise, such as key
                              already in database.
        """
        return (self.db.putVal(db=self.sdb,
                               key=self._tokey(keys),
                               val=self._ser(val)))


    def pin(self, keys: Union[str, Iterable], val: Union[bytes, str]):
        """
        Pins (sets) val at key made from keys. Overwrites.

        Parameters:
            keys (tuple): of key strs to be combined in order to form key
            val (bytes): value

        Returns:
            result (bool): True If successful. False otherwise.
        """
        return (self.db.setVal(db=self.sdb,
                               key=self._tokey(keys),
                               val=self._ser(val)))


    def get(self, keys: Union[str, Iterable]):
        """
        Gets val at keys

        Parameters:
            keys (tuple): of key strs to be combined in order to form key

        Returns:
            data (str):  decoded as utf-8
            None if no entry at keys

        Usage:
            Use walrus operator to catch and raise missing entry
            if (data := mydb.get(keys)) is None:
                raise ExceptionHere
            use data here

        """
        val = self.db.getVal(db=self.sdb, key=self._tokey(keys))
        return (self._des(val) if val is not None else None)


    def rem(self, keys: Union[str, Iterable]):
        """
        Removes entry at keys

        Parameters:
            keys (tuple): of key strs to be combined in order to form key

        Returns:
           result (bool): True if key exists so delete successful. False otherwise
        """
        return(self.db.delVal(db=self.sdb, key=self._tokey(keys)))


class OnSuberBase(SuberBase):
    """
    Subclass of SuberBase that adds methods for keys with  exposed key part suffix
    that is 32 byte serializaton of monotonically increasing ordinal number on
    such as sn or fn.
    Each key consistes of top key joined with .sep to ordinal suffix
    Works with dupsort==True or False

    """

    def __init__(self, *pa, **kwa):
        """
        Inherited Parameters:
            db (dbing.LMDBer): base db
            subkey (str):  LMDB sub database key
            dupsort (bool): True means enable duplicates at each key
                               False (default) means do not enable duplicates at
                               each key. Default False
            sep (str): separator to convert keys iterator to key bytes for db key
                       Default '.'
            verify (bool): True means reverify when ._des from db when applicable
                           False means do not reverify. Default False
        """
        super(OnSuberBase, self).__init__(*pa, **kwa)


    def putOn(self, keys: str | bytes | memoryview, on: int=0,
                    val: str | bytes | memoryview=''):
        """
        Returns
            result (bool): True if onkey made from key+sep+serialized on is
                               not found in database so value is written
                               idempotently.
                           False otherwise

        Parameters:
            keys (str | bytes | memoryview | Iterable): keys as prefix to be
                combined with serialized on suffix and sep to form onkey
            on (int): ordinal number used with onKey(key ,on) to form key.
            val (str | bytes | memoryview): serialization
        """
        return (self.db.putOnVal(db=self.sdb,
                                 key=self._tokey(keys),
                                 on=on,
                                 val=self._ser(val),
                                 sep=self.sep.encode()))

    def pinOn(self, keys: str | bytes | memoryview, on: int=0,
                    val: str | bytes | memoryview=''):
        """
        Returns
            result (bool): True if value is written or overwritten at onkey
                           False otherwise

        Parameters:
            keys (str | bytes | memoryview | Iterable): keys as prefix to be
                combined with serialized on suffix and sep to form onkey
            on (int): ordinal number used with onKey(key ,on) to form key.
            val (str | bytes | memoryview): serialization
        """
        return (self.db.setOnVal(db=self.sdb,
                                 key=self._tokey(keys),
                                 on=on,
                                 val=self._ser(val),
                                 sep=self.sep.encode()))


    def appendOn(self, keys: str | bytes | memoryview,
                       val: str | bytes | memoryview):
        """
        Returns:
            on (int): ordinal number of newly appended val

        Parameters:
            keys (str | bytes | memoryview | Iterable): top keys as prefix to be
                combined with serialized on suffix and sep to form key
            val (str | bytes | memoryview): serialization
        """
        return (self.db.appendOnVal(db=self.sdb,
                                       key=self._tokey(keys),
                                       val=self._ser(val),
                                       sep=self.sep.encode()))


    def getOn(self, keys: str | bytes | memoryview, on: int=0):
        """
        Returns
            val (str): serialization at onkey if any
                       None if no entry at onkey

        Parameters:
            keys (str | bytes | memoryview | Iterable): keys as prefix to be
                combined with serialized on suffix and sep to form onkey
            on (int): ordinal number used with onKey(key ,on) to form key.
        """
        val = self.db.getOnVal(db=self.sdb,
                                key=self._tokey(keys),
                                on=on,
                                sep=self.sep.encode())
        return (self._des(val) if val is not None else None)



    def remOn(self, keys: str | bytes | memoryview, on: int=0):
        """
        Returns
            result (bool): True if onkey made from key+sep+serialized on is
                               found in database so value is removed
                               Removes all duplicates if any at onkey.
                           False otherwise

        Parameters:
            keys (str | bytes | memoryview | Iterable): keys as prefix to be
                combined with serialized on suffix and sep to form onkey
            on (int): ordinal number used with onKey(key ,on) to form key.
        """
        return (self.db.delOnVal(db=self.sdb,
                                     key=self._tokey(keys),
                                     on=on,
                                     sep=self.sep.encode()))


    def cntOn(self, keys: str | bytes | memoryview = "", on: int=0):
        """
        Returns
            cnt (int): count of of all ordinal suffix keyed vals with same
                key prefix but different on in onkey in db starting at ordinal
                number on where key is formed with onKey(key,on). Count at
                each onkey includes duplicates if any.


        Parameters:
            keys (str | bytes | memoryview | Iterable): top keys as prefix to be
                combined with serialized on suffix and sep to form top key
                When keys is empty then counts whole database including
                duplicates if any.
            on (int): ordinal number used with onKey(key,on) to form key.
        """
        return (self.db.cntOnVals(db=self.sdb,
                                     key=self._tokey(keys),
                                     on=on,
                                     sep=self.sep.encode()))


    def getOnIter(self, keys: str|bytes|memoryview|Iterable = "", on: int=0):
        """
        Returns:
            items (Iterator[bytes]): of val with same key but increments of
                                on >= on i.e. all key.on beginning with on

        Parameters:
            keys (str | bytes | memoryview | iterator): keys as prefix to be
                combined with serialized on suffix and sep to form actual key
                When keys is empty then retrieves whole database including
                duplicates if any
            on (int): ordinal number used with onKey(pre,on) to form key at at
                      which to initiate retrieval
            sep (bytes): separator character for split
        """
        for val in (self.db.getOnValIter(db=self.sdb,
                        key=self._tokey(keys), on=on, sep=self.sep.encode())):
            yield (self._des(val))


    def getOnItemIter(self, keys: str|bytes|memoryview|Iterable = "", on: int=0):
        """
        Returns:
            items (Iterator[(key, on, val)]): triples of key, on, val with same
                key but increments of on >= on i.e. all key.on beginning with on

        Parameters:
            keys (str | bytes | memoryview | iterator): keys as prefix to be
                combined with serialized on suffix and sep to form actual key
                When keys is empty then retrieves whole database including
                duplicates if any
            on (int): ordinal number used with onKey(pre,on) to form key at at
                      which to initiate retrieval
            sep (bytes): separator character for split
        """
        for keys, on, val in (self.db.getOnItemIter(db=self.sdb,
                        key=self._tokey(keys), on=on, sep=self.sep.encode())):
            yield (self._tokeys(keys), on, self._des(val))



class OnSuber(OnSuberBase, Suber):
    """
    Subclass of OnSuberBase andSuber that adds methods for keys with ordinal
    numbered suffixes.
    Each key consistes of pre joined with .sep to ordinal suffix

    Assumes dupsort==False

    """

    def __init__(self, *pa, **kwa):
        """
        Inherited Parameters:
            db (dbing.LMDBer): base db
            subkey (str):  LMDB sub database key
            dupsort (bool): True means enable duplicates at each key
                               False (default) means do not enable duplicates at
                               each key. Set to False
            sep (str): separator to convert keys iterator to key bytes for db key
                       default is self.Sep == '.'
            verify (bool): True means reverify when ._des from db when applicable
                           False means do not reverify. Default False
        """
        super(OnSuber, self).__init__(*pa, **kwa)


class B64SuberBase(SuberBase):
    """
    Base Class whose values are Iterables of Base64 str or bytes that are stored
    in db as .sep joined Base64 bytes. Separator character must not be valid
    Base64 character so the split will work unambiguously.

    Automatically joins and splits along separator to Iterable (tuple) of Base64

     Attributes:
        db (dbing.LMDBer): base LMDB db
        sdb (lmdb._Database): instance of lmdb named sub db for this Suber
        sep (str): separator for combining keys tuple of strs into key bytes
    """

    def __init__(self, *pa, **kwa):
        """
        Inherited Parameters:
            db (dbing.LMDBer): base db
            subkey (str):  LMDB sub database key
            dupsort (bool): True means enable duplicates at each key
                               False (default) means do not enable duplicates at
                               each key
            sep (str): separator to convert keys iterator to key bytes for db key
                       default is self.Sep == '.'
                       Must not be Base64 character.
            verify (bool): True means reverify when ._des from db when applicable
                           False means do not reverify. Default False

        """
        super(B64SuberBase, self).__init__(*pa, **kwa)
        if Reb64.match(self.sep.encode()):
            raise ValueError("Invalid sep={self.sep}, must not be Base64 char.")


    def _toval(self, vals: str|bytes|memoryview|Iterable[str|bytes|memoryview]):
        """
        Converts vals to val bytes with proper separators and returns val bytes.
        If vals is already str or bytes or memoryview then returns val bytes.
        Else If vals is iterable (non-str) of strs or bytes or memoryview then
        joins with .sep and converts to val bytes and returns.

        Returns:
           val (bytes): each element of vals is joined by .sep.

        Parameters:
           vals (str | bytes | memoryview | Iterable[str | bytes]): db val or
                        Iterable of (str | bytes | memoryview) to form val.
                        Note, join of bytes sep works with memoryview.

        """
        if hasattr(vals, "encode"):  # str
            val = vals.encode("utf-8")
            if not (Reb64.match(val)):
                raise ValueError(f"Non Base64 {val=}.")
            return val
        if isinstance(vals, memoryview):  # memoryview of bytes
            val = bytes(vals)  # return bytes
            if not (Reb64.match(val)):
                raise ValueError(f"Non Base64 {val=}.")
            return val
        elif hasattr(vals, "decode"): # bytes
            val = vals
            if not (Reb64.match(val)):
                raise ValueError(f"Non Base64 {val=}.")
            return val
        vals = tuple(v.encode() if hasattr(v, "encode") else v for v in vals)  # make bytes
        for val in vals:
            if not (Reb64.match(val)):
                raise ValueError(f"Non Base64 {val=}.")
        return (self.sep.encode().join(vals))

        #return (self.sep.join(val.decode() if hasattr(val, "decode") else val
                              #for val in vals).encode("utf-8"))


    def _tovals(self, val: bytes | memoryview):
        """
        Converts val bytes to vals tuple of strs by decoding and then splitting
        at separator .sep.

        Returns:
           vals (tuple[str]): makes tuple by splitting val at .sep

        Parameters:
           val (bytes | memoryview): db Base64 val.

        """
        if isinstance(val, memoryview):  # memoryview of bytes
            val = bytes(val)
        if hasattr(val, "decode"):  # bytes
            val = val.decode("utf-8")  # convert to str
        return tuple(val.split(self.sep))


    def _ser(self, val: Union[Iterable, str, bytes]):
        """
        Serialize val to bytes to store in db
        When val is Iterable then joins each elements with .sep returns val bytes

        Returns:
           val (bytes): .sep join of each Base64 bytes in val

        Parameters:
           val (Union[Iterable, bytes]): of Base64 bytes

        """
        if not nonStringIterable(val):  # not iterable
            val = (val, )  # make iterable
        return (self._toval(val))


    def _des(self, val: memoryview | bytes):
        """
        Converts val bytes to vals tuple of subclass instances by deserializing
        .qb64b  concatenation in order of each instance in .klas

        Returns:
           vals (tuple): subclass instances

        Parameters:
           val (Union[bytes, memoryview]):  of concatenation of .qb64b

        """
        return self._tovals(val)


class B64Suber(B64SuberBase, Suber):
    """
    Subclass of B64SuberBase and Suber that serializes and deserializes values
    as .sep joined strings of Base64 components.

    .sep must not be Base64 character.

    Each key consistes of pre joined with .sep to ordinal suffix

    Assumes dupsort==False

    """

    def __init__(self, *pa, **kwa):
        """
        Inherited Parameters:
            db (dbing.LMDBer): base db
            subkey (str):  LMDB sub database key
            dupsort (bool): True means enable duplicates at each key
                               False (default) means do not enable duplicates at
                               each key. Set to False
            sep (str): separator to convert keys iterator to key bytes for db key
                       default is self.Sep == '.'
                       Must not be Base64 character.
            verify (bool): True means reverify when ._des from db when applicable
                           False means do not reverify. Default False
        """
        super(B64Suber, self).__init__(*pa, **kwa)



class CesrSuberBase(SuberBase):
    """
    Sub class of SuberBase where data is CESR encode/decode ducktyped subclass
    instance such as Matter, Indexer, Counter with .qb64b property when provided
    as fully qualified serialization
    Automatically serializes and deserializes from qb64b to/from CESR instance

    """

    def __init__(self, *pa, klas: Type[coring.Matter] = coring.Matter, **kwa):
        """
        Inherited Parameters:
            db (dbing.LMDBer): base db
            subkey (str):  LMDB sub database key
            dupsort (bool): True means enable duplicates at each key
                               False (default) means do not enable duplicates at
                               each key
            sep (str): separator to convert keys iterator to key bytes for db key
                       default is self.Sep == '.'
            verify (bool): True means reverify when ._des from db when applicable
                           False means do not reverify. Default False
        Parameters:
            klas (Type[coring.Matter]): Class reference to subclass of Matter or
                Indexer or Counter or any ducktyped class of Matter

        """
        super(CesrSuberBase, self).__init__(*pa, **kwa)
        self.klas = klas


    def _ser(self, val: coring.Matter):
        """
        Serialize value to bytes to store in db
        Parameters:
            val (coring.Matter): instance Matter ducktype with .qb64b attribute
        """
        return val.qb64b


    def _des(self, val: memoryview | bytes):
        """
        Deserialize val to str
        Parameters:
            val (memoryview | bytes): convertable to coring.matter
        """
        if isinstance(val, memoryview):  # memoryview is always bytes
            val = bytes(val)  # convert to bytes
        return self.klas(qb64b=val)  # qb64b parameter accepts str


class CesrSuber(CesrSuberBase, Suber):
    """
    Sub class of Suber where data is CESR encode/decode ducktyped subclass
    instance such as Matter, Indexer, Counter with .qb64b property when provided
    as fully qualified serialization.
    Extends Suber to support val that are ducktyped CESR serializable .qb64 .qb64b
    subclasses such as coring.Matter, coring.Indexer, coring.Counter.
    Automatically serializes and deserializes from qb64b to/from CESR instances

    """

    def __init__(self, *pa, **kwa):
        """
        Inherited Parameters:
            db (dbing.LMDBer): base db
            subkey (str):  LMDB sub database key
            dupsort (bool): True means enable duplicates at each key
                               False (default) means do not enable duplicates at
                               each key
            sep (str): separator to convert keys iterator to key bytes for db key
                       default is self.Sep == '.'
            verify (bool): True means reverify when ._des from db when applicable
                           False means do not reverify. Default False
            klas (Type[coring.Matter]): Class reference to subclass of Matter or
                Indexer or Counter or any ducktyped class of Matter
        """
        super(CesrSuber, self).__init__(*pa, **kwa)


class CesrOnSuber(CesrSuberBase, OnSuberBase, Suber):
    """
    Subclass of CesrSuberBase, OnSuberBase, and Suber that adds methods for
    keys with ordinal numbered suffixes and values that are Cesr serializations
    of Matter subclass ducktypes.

    Each key consistes of pre joined with .sep to ordinal suffix

    Assumes dupsort==False
    """

    def __init__(self, *pa, **kwa):
        """
        Inherited Parameters:
            db (dbing.LMDBer): base db
            subkey (str):  LMDB sub database key
            dupsort (bool): True means enable duplicates at each key
                               False (default) means do not enable duplicates at
                               each key. Set to False
            sep (str): separator to convert keys iterator to key bytes for db key
                       default is self.Sep == '.'
            verify (bool): True means reverify when ._des from db when applicable
                           False means do not reverify. Default False
        """
        super(CesrOnSuber, self).__init__(*pa, **kwa)


class CatCesrSuberBase(CesrSuberBase):
    """
    Base Class whose values stored in db are a concatenation of the  .qb64b property
    from one or more  subclass instances (qb64b is bytes of fully qualified
    serialization) that support CESR encode/decode ducktyped subclass instance
    such as Matter, Indexer, Counter
    Automatically serializes and deserializes from qb64b to/from CESR instances

     Attributes:
        db (dbing.LMDBer): base LMDB db
        sdb (lmdb._Database): instance of lmdb named sub db for this Suber
        sep (str): separator for combining keys tuple of strs into key bytes
        klas (Iterable): of Class references to subclasses of CESR compatible
                , each of to Type[coring.Matter etc]
    """

    def __init__(self, *pa, klas: Iterable = None, **kwa):
        """
        Inherited Parameters:
            db (dbing.LMDBer): base db
            subkey (str):  LMDB sub database key
            dupsort (bool): True means enable duplicates at each key
                               False (default) means do not enable duplicates at
                               each key
            sep (str): separator to convert keys iterator to key bytes for db key
                       default is self.Sep == '.'
            verify (bool): True means reverify when ._des from db when applicable
                           False means do not reverify. Default False
            klas (Type[coring.Matter]): Class reference to subclass of Matter or
                Indexer or Counter or any ducktyped class of Matter

        """
        if klas is None:
            klas = (coring.Matter, )  # set default to tuple of single Matter
        if not nonStringIterable(klas):  # not iterable
            klas = (klas, )  # make it so
        super(CatCesrSuberBase, self).__init__(*pa, klas=klas, **kwa)


    def _ser(self, val: Union[Iterable, coring.Matter]):
        """
        Serialize val to bytes to store in db
        Concatenates .qb64b of each instance in val and returns val bytes

        Returns:
           cat (bytes): concatenation of .qb64b of each object instance in vals

        Parameters:
           val (Union[Iterable, coring.Matter]): of subclass instances.

        """
        if not nonStringIterable(val):  # not iterable
            val = (val, )  # make iterable
        return (b''.join(obj.qb64b for obj in val))


    def _des(self, val: memoryview | bytes | bytearray):
        """
        Converts val bytes to vals tuple of subclass instances by deserializing
        .qb64b  concatenation in order of each instance in .klas

        Returns:
           vals (tuple): subclass instances

        Parameters:
           val (Union[bytes, memoryview]):  of concatenation of .qb64b

        """
        if not isinstance(val, bytearray):  # is memoryview or bytes
            val = bytearray(val)  # convert so may strip
        return tuple(klas(qb64b=val, strip=True) for klas in self.klas)


class CatCesrSuber(CatCesrSuberBase, Suber):
    """
    Class whose values stored in db are a concatenation of the  .qb64b property
    from one or more  subclass instances (qb64b is bytes of fully qualified
    serialization) that support CESR encode/decode ducktyped subclass instance
    such as Matter, Indexer, Counter
    Automatically serializes and deserializes from qb64b to/from CESR instances

    Attributes:
        db (dbing.LMDBer): base LMDB db
        sdb (lmdb._Database): instance of lmdb named sub db for this Suber
        sep (str): separator for combining keys tuple of strs into key bytes
        klas (Iterable): of Class references to subclasses of CESR compatible
                , each of to Type[coring.Matter etc]
    """

    def __init__(self, *pa, **kwa):
        """
        Inherited Parameters:
            db (dbing.LMDBer): base db
            subkey (str):  LMDB sub database key
            dupsort (bool): True means enable duplicates at each key
                               False (default) means do not enable duplicates at
                               each key
            sep (str): separator to convert keys iterator to key bytes for db key
                       default is self.Sep == '.'
            verify (bool): True means reverify when ._des from db when applicable
                           False means do not reverify. Default False
            klas (Type[coring.Matter]): Class reference to subclass of Matter or
                Indexer or Counter or any ducktyped class of Matter

        """
        super(CatCesrSuber, self).__init__(*pa, **kwa)


class IoSetSuber(SuberBase):
    """
    Insertion Ordered Set Suber factory class that supports
    a set of distinct entries at a given effective database key but with
    dupsort==False. Effective data model is that there are multiple values in a
    set of values where every member of the set has the same key (duplicate key).
    The set of values is an ordered set using insertion order. Any given value
    may appear only once in the set (not a list).

    This works similarly to the IO value duplicates for the LMDBer class with a
    sub db  of LMDB (dupsort==True) but without its size limitation of 511 bytes
    for each value when dupsort==True.
    Here the key is augmented with a hidden numbered suffix that provides a
    an ordered set of values at each effective key (duplicate key). The suffix
    is appended and stripped transparently. The set of multiple items with
    duplicate keys are retrieved in insertion order when iterating or as a list
    of the set elements.

    Attributes:
        db (dbing.LMDBer): base LMDB db
        sdb (lmdb._Database): instance of lmdb named sub db for this Suber
        sep (str): separator for combining keys tuple of strs into key bytes
    """
    def __init__(self, db: dbing.LMDBer, *,
                       subkey: str='docs.',
                       dupsort: bool=False, **kwa):
        """
        Inherited Parameters:
            db (dbing.LMDBer): base db
            subkey (str):  LMDB sub database key
            dupsort (bool): True means enable duplicates at each key
                               False (default) means do not enable duplicates at
                               each key
            sep (str): separator to convert keys iterator to key bytes for db key
                       default is self.Sep == '.'
            verify (bool): True means reverify when ._des from db when applicable
                           False means do not reverify. Default False
            klas (Type[coring.Matter]): Class reference to subclass of Matter or
                Indexer or Counter or any ducktyped class of Matter
        """
        super(IoSetSuber, self).__init__(db=db, subkey=subkey, dupsort=False, **kwa)


    def put(self, keys: str | bytes | memoryview | Iterable,
                  vals: str | bytes | memoryview | Iterable):
        """
        Puts all vals at effective key made from keys and hidden ordinal suffix.
        that are not already in set of vals at key. Does not overwrite.

        Parameters:
            keys (str | bytes | memoryview | Iterable): of key parts to be
                    combined in order to form key
            vals (str | bytes | memoryview | Iterable): of str serializations

        Returns:
            result (bool): True If successful, False otherwise.

        """
        if not nonStringIterable(vals):  # not iterable
            vals = (vals, )  # make iterable
        return (self.db.putIoSetVals(db=self.sdb,
                                     key=self._tokey(keys),
                                     vals=[self._ser(val) for val in vals],
                                     sep=self.sep))


    def add(self, keys: str | bytes | memoryview | Iterable,
            val: str | bytes | memoryview):
        """
        Add val idempotently to vals at effective key made from keys and hidden
        ordinal suffix. Idempotent means that added value is not already in set
        of vals at key. Does not overwrite or add same value at same key more
        than once.

        Parameters:
            keys (str | bytes | memoryview | Iterable): of key parts to be
                    combined in order to form key
            val (str | bytes | memoryview): serialization

        Returns:
            result (bool): True means unique value added among duplications,
                            False means duplicate of same value already exists.

        """
        return (self.db.addIoSetVal(db=self.sdb,
                                    key=self._tokey(keys),
                                    val=self._ser(val),
                                    sep=self.sep))


    def pin(self, keys: str | bytes | memoryview | Iterable,
                  vals: str | bytes | memoryview | Iterable):
        """
        Pins (sets) vals at effective key made from keys and hidden ordinal suffix.
        Overwrites. Removes all pre-existing vals that share same effective keys
        and replaces them with vals

        Parameters:
            keys (Iterable): of key strs to be combined in order to form key
            vals (Iterable): str serializations

        Returns:
            result (bool): True If successful, False otherwise.

        """
        if not nonStringIterable(vals):  # not iterable
            vals = (vals, )  # make iterable
        return (self.db.setIoSetVals(db=self.sdb,
                                     key=self._tokey(keys),
                                     vals=[self._ser(val) for val in vals],
                                     sep=self.sep))


    def get(self, keys: str | bytes | memoryview | Iterable):
        """
        Gets vals set list at key made from effective keys

        Parameters:
            keys (Iterable): of key strs to be combined in order to form key

        Returns:
            vals (Iterable):  each item in list is str
                          empty list if no entry at keys

        """
        return ([self._des(val) for val in
                    self.db.getIoSetValsIter(db=self.sdb,
                                             key=self._tokey(keys),
                                             sep=self.sep)])


    def getIter(self, keys: str | bytes | memoryview | Iterable):
        """
        Gets vals iterator at effecive key made from keys and hidden ordinal suffix.
        All vals in set of vals that share same effecive key are retrieved in
        insertion order.

        Parameters:
            keys (Iterable): of key strs to be combined in order to form key

        Returns:
            vals (Iterator):  str values. Raises StopIteration when done

        """
        for val in self.db.getIoSetValsIter(db=self.sdb,
                                            key=self._tokey(keys),
                                            sep=self.sep):
            yield self._des(val)


    def getLast(self, keys: str | bytes | memoryview | Iterable):
        """
        Gets last val inserted at effecive key made from keys and hidden ordinal
        suffix.

        Parameters:
            keys (Iterable): of key strs to be combined in order to form key

        Returns:
            val (str):  value str, None if no entry at keys

        """
        val = self.db.getIoSetValLast(db=self.sdb, key=self._tokey(keys))
        return (self._des(val) if val is not None else val)



    def rem(self, keys: str | bytes | memoryview | Iterable,
                   val: str | bytes | memoryview = b''):
        """
        Removes entry at effective key made from keys and hidden ordinal suffix
        that matches val if any. Otherwise deletes all values at effective key.

        Parameters:
            keys (Iterable): of key strs to be combined in order to form key
            val (str):  value at key to delete. Subclass ._ser method may
                        accept different value types
                        if val is empty then remove all values at key

        Returns:
           result (bool): True if effective key with val exists so rem successful.
                           False otherwise

        """
        if val:
            return self.db.delIoSetVal(db=self.sdb,
                                       key=self._tokey(keys),
                                       val=self._ser(val),
                                       sep=self.sep)
        else:
            return self.db.delIoSetVals(db=self.sdb,
                                       key=self._tokey(keys),
                                       sep=self.sep)


    def cnt(self, keys: str | bytes | memoryview | Iterable):
        """
        Return count of  values at effective key made from keys and hidden ordinal
        suffix. Zero otherwise

        Parameters:
            keys (Iterable): of key strs to be combined in order to form key
        """
        return (self.db.cntIoSetVals(db=self.sdb,
                                     key=self._tokey(keys),
                                     sep=self.sep))


    def getItemIter(self, keys: str | bytes | memoryview | Iterable = "",
                    *, topive=False):
        """
        Return iterator over all the items in top branch defined by keys where
        keys may be truncation of full branch.

        Returns:
            items (Iterator): of (key, val) tuples over the all the items in
            subdb whose effective key startswith key made from keys.
            Keys may be keyspace prefix in order to return branches of key space.
            When keys is empty then returns all items in subdb.
            Returned key in each item has ordinal suffix removed.

        Parameters:
            keys (Iterable): tuple of bytes or strs that may be a truncation of
                a full keys tuple in  in order to address all the items from
                multiple branches of the key space. If keys is empty then gets
                all items in database.
                Either append "" to end of keys Iterable to ensure get properly
                separated top branch key or use top=True.
                In Python str.startswith('') always returns True so if branch
                key is empty string it matches all keys in db with startswith.

            topive (bool): True means treat as partial key tuple from top branch of
                key space given by partial keys. Resultant key ends in .sep
                character.
                False means treat as full branch in key space. Resultant key
                does not end in .sep character.
                When last item in keys is empty str then will treat as
                partial ending in sep regardless of top value

        """
        for key, val in self.db.getTopIoSetItemIter(db=self.sdb,
                top=self._tokey(keys, topive=topive), sep=self.sep.encode()):
            yield (self._tokeys(key), self._des(val))


class CesrIoSetSuber(CesrSuberBase, IoSetSuber):
    """
    Subclass of CesrSuber and IoSetSuber.
    Sub class of Suber where data is CESR encode/decode ducktyped subclass
    instance such as Matter, Indexer, Counter with .qb64b property when provided
    as fully qualified serialization
    Automatically serializes and deserializes from qb64b to/from CESR instances

    Extends IoSetSuber with mixin methods ._ser and ._des from CesrSuberBase
    so that all IoSetSuber methods now work with CESR subclass for each val.

    IoSetSuber stores at each effective key a set of distinct values that
    share that same effective key where each member of the set is retrieved in
    insertion order (dupsort==False)
    The methods allows an Iterable (set valued) of Iterables of Matter subclass
    instances to be stored at a given effective key in insertion order.

    Actual keys include a hidden ordinal key suffix that tracks insertion order.
    The suffix is appended and stripped transparently from the keys. The set of
    items with duplicate effective keys are retrieved in insertion order when
    iterating or as a list of the set elements. The actual iokey for any item
    includes the ordinal suffix.

     Attributes:
        db (dbing.LMDBer): base LMDB db
        sdb (lmdb._Database): instance of lmdb named sub db for this Suber
        sep (str): separator for combining keys tuple of strs into key bytes
        klas (Iterable): of Class references to subclasses of CESR compatible
                , each of to Type[coring.Matter etc]
    """

    def __init__(self, *pa, **kwa):
        """
        Inherited Parameters:
            db (dbing.LMDBer): base db
            subkey (str):  LMDB sub database key
            dupsort (bool): True means enable duplicates at each key
                               False (default) means do not enable duplicates at
                               each key
            sep (str): separator to convert keys iterator to key bytes for db key
                       default is self.Sep == '.'
            verify (bool): True means reverify when ._des from db when applicable
                           False means do not reverify. Default False
            klas (Type[coring.Matter]): Class reference to subclass of Matter or
                Indexer or Counter or any ducktyped class of Matter

        """
        super(CesrIoSetSuber, self).__init__(*pa, **kwa)



class CatCesrIoSetSuber(CatCesrSuberBase, IoSetSuber):
    """
    Sub class of CatSuberBase and IoSetSuber where values stored in db are a
    concatenation of .qb64b property from one or more Cesr compatible subclass
    instances that automatically serializes and deserializes to/from qb64b .
    (qb64b is bytes of fully qualified serialization).

    Extends IoSetSuber with mixin methods ._ser and ._des from CatSuberBase
    so that all IoSetSuber methods now work with an Iterable of CESR subclass
    for each val.

    IoSetSuber stores at each effective key a set of distinct values that
    share that same effective key where each member of the set is retrieved in
    insertion order (dupsort==False)
    The methods allows an Iterable (set valued) of Iterables of Matter subclass
    instances to be stored at a given effective key in insertion order.

    Actual keys include a hidden ordinal key suffix that tracks insertion order.
    The suffix is appended and stripped transparently from the keys. The set of
    items with duplicate effective keys are retrieved in insertion order when
    iterating or as a list of the set elements. The actual iokey for any item
    includes the ordinal suffix.

    Attributes:
        db (dbing.LMDBer): base LMDB db
        sdb (lmdb._Database): instance of lmdb named sub db for this Suber
        sep (str): separator for combining keys tuple of strs into key bytes
        klas (Iterable): of Class references to subclasses of Matter, each
                of to Type[coring.Matter]


    """
    def __init__(self, *pa, **kwa):
        """
        Inherited Parameters:
            db (dbing.LMDBer): base db
            subkey (str):  LMDB sub database key
            dupsort (bool): True means enable duplicates at each key
                               False (default) means do not enable duplicates at
                               each key
            sep (str): separator to convert keys iterator to key bytes for db key
                       default is self.Sep == '.'
            verify (bool): True means reverify when ._des from db when applicable
                           False means do not reverify. Default False
            klas (Type[coring.Matter]): Class reference to subclass of Matter or
                Indexer or Counter or any ducktyped class of Matter

        """
        super(CatCesrIoSetSuber, self).__init__(*pa, **kwa)



class SignerSuber(CesrSuber):
    """
    Sub class of CesrSuber where data is Signer subclass instance .qb64b propery
    which is a fully qualified serialization and uses the key which is the qb64b
    of the signer.verfer to get the transferable property of the verfer
    Automatically serializes and deserializes from qb64b to/from Signer instances

    Assumes that last or only element of db key from keys for all entries is the qb64
    of a public key for the associated Verfer instance. This allows returned
    Signer instance to have its .transferable property set correctly.
    """

    def __init__(self, *pa, klas: Type[core.Signer] = core.Signer, **kwa):
        """
        Parameters:
            db (dbing.LMDBer): base db
            subkey (str):  LMDB sub database key
            klas (Type[coring.Matter]): Class reference to subclass of Matter
        """
        if not (issubclass(klas, core.Signer)):
            raise ValueError("Invalid klas type={}, expected {}."
                             "".format(klas, core.Signer))
        super(SignerSuber, self).__init__(*pa, **kwa)
        self.klas = klas


    def get(self, keys: Union[str, Iterable]):
        """
        Gets Signer instance at keys

        Returns:
            val (Signer):  transferable determined by key which is verfer
            None if no entry at keys

        Parameters:
            keys (Union[str, iterable]): key strs to be combined in order to
                form key. Last element of keys is verkey used to determin
                .transferable for Signer

        Usage:
            Use walrus operator to catch and raise missing entry
            if (signer := mydb.get(keys)) is None:
                raise ExceptionHere
            use signer here

        """
        key = self._tokey(keys)  # keys maybe string or tuple
        val = self.db.getVal(db=self.sdb, key=key)
        keys = self._tokeys(key)  # verkey is last split if any
        verfer = coring.Verfer(qb64b=keys[-1])  # last split
        return (self.klas(qb64b=bytes(val), transferable=verfer.transferable)
                if val is not None else None)


    def getItemIter(self, keys: str | bytes | memoryview | Iterable = "",
                    *, topive=False):
        """
        Returns:
            iterator (Iteratore: tuple (key, val) over the all the items in
            subdb whose key startswith key made from keys. Keys may be keyspace
            prefix to return branches of key space. When keys is empty then
            returns all items in subdb

        Parameters:
            keys (Iterable): tuple of bytes or strs that may be a truncation of
                a full keys tuple in  in order to get all the items from
                multiple branches of the key space. If keys is empty then gets
                all items in database.
            topive (bool): True means treat as partial key tuple from top branch of
                key space given by partial keys. Resultant key ends in .sep
                character.
                False means treat as full branch in key space. Resultant key
                does not end in .sep character.
                When last item in keys is empty str then will treat as
                partial ending in sep regardless of top value

        """
        for key, val in self.db.getTopItemIter(db=self.sdb,
                                        top=self._tokey(keys, topive=topive)):
            ikeys = self._tokeys(key)  # verkey is last split if any
            verfer = coring.Verfer(qb64b=ikeys[-1])   # last split
            yield (ikeys, self.klas(qb64b=bytes(val),
                                   transferable=verfer.transferable))


class CryptSignerSuber(SignerSuber):
    """
    Sub class of SignerSuber where data is Signer subclass instance .qb64b property
    that has been encrypted if encrypter provided.
    which is a fully qualified serialization and uses the key which is the qb64b
    of the signer.verfer to get the transferable property of the verfer
    Automatically serializes and deserializes from qb64b to/from Signer instances

    Assumes that last or only element of db key from keys for all entries is the qb64
    of a public key for the associated Verfer instance. This allows returned
    Signer instance to have its .transferable property set correctly.
    """

    def put(self, keys: Union[str, Iterable], val: coring.Matter,
            encrypter: core.Encrypter = None):
        """
        Puts qb64 of Matter instance val at key made from keys. Does not overwrite
        If encrypter provided then encrypts first

        Parameters:
            keys (tuple): of key strs to be combined in order to form key
            val (Signer): instance of self.klas
            encrypter (core.Encrypter): optional

        Returns:
            result (bool): True If successful, False otherwise, such as key
                              already in database.
        """
        if encrypter:
            val = encrypter.encrypt(prim=val)  # returns Cipher instance
        return (self.db.putVal(db=self.sdb,
                               key=self._tokey(keys),
                               val=val.qb64b))


    def pin(self, keys: Union[str, Iterable], val: coring.Matter,
            encrypter: core.Encrypter = None):
        """
        Pins (sets) qb64 of Matter instance val at key made from keys. Overwrites.
        If encrypter provided then encrypts first

        Parameters:
            keys (tuple): of key strs to be combined in order to form key
            val (Signer): instance of self.klas
            encrypter (core.Encrypter): optional

        Returns:
            result (bool): True If successful. False otherwise.
        """
        if encrypter:
            val = encrypter.encrypt(prim=val)  # returns Cipher instance
        return (self.db.setVal(db=self.sdb,
                               key=self._tokey(keys),
                               val=val.qb64b))



    def get(self, keys: Union[str, Iterable], decrypter: core.Decrypter = None):
        """
        Gets Signer instance at keys. If decrypter then assumes value in db was
        encrypted and so decrypts value in db before converting to Signer.


        Returns:
            val (Signer):  transferable determined by key which is verfer
            None if no entry at keys

        Parameters:
            keys (Union[str, iterable]): key strs to be combined in order to
                form key. Last element of keys is verkey used to determin
                .transferable for Signer
            decrypter (core.Decrypter): optional. If provided assumes value in
                db was encrypted and so decrypts before converting to Signer.

        Usage:
            Use walrus operator to catch and raise missing entry
            if (signer := mydb.get(keys)) is None:
                raise ExceptionHere
            use signer here

        """
        key = self._tokey(keys)  # keys maybe string or tuple
        val = self.db.getVal(db=self.sdb, key=key)
        if val is None:
            return None
        keys = self._tokeys(key)  # verkey is last split if any
        verfer = coring.Verfer(qb64b=keys[-1])  # last split
        if decrypter:
            return (decrypter.decrypt(qb64=bytes(val),
                                      transferable=verfer.transferable))
        return (self.klas(qb64b=bytes(val), transferable=verfer.transferable))



    def getItemIter(self, keys: str|bytes|memoryview|Iterable= "",
                       decrypter: core.Decrypter = None, *, topive=False):
        """
        Returns:
            items (Iterator): of tuples (key, val) over the all the items in
            subdb whose key startswith key made from keys. Keys may be keyspace
            prefix to return branches of key space. When keys is empty then
            returns all items in subdb

        decrypter (core.Decrypter): optional. If provided assumes value in
                db was encrypted and so decrypts before converting to Signer.

        Parameters:
            keys (Iterable): tuple of bytes or strs that may be a truncation of
                a full keys tuple in  in order to get all the items from
                multiple branches of the key space. If keys is empty then gets
                all items in database.
            topive (bool): True means treat as partial key tuple from top branch of
                key space given by partial keys. Resultant key ends in .sep
                character.
                False means treat as full branch in key space. Resultant key
                does not end in .sep character.
                When last item in keys is empty str then will treat as
                partial ending in sep regardless of top value

        """
        for key, val in self.db.getTopItemIter(db=self.sdb,
                                        top=self._tokey(keys, topive=topive)):
            ikeys = self._tokeys(key)  # verkey is last split if any
            verfer = coring.Verfer(qb64b=ikeys[-1])   # last split
            if decrypter:
                yield (ikeys, decrypter.decrypt(qb64=bytes(val),
                                            transferable=verfer.transferable))
            else:
                yield (ikeys, self.klas(qb64b=bytes(val),
                                            transferable=verfer.transferable))

class SerderSuberBase(SuberBase):
    """
    Sub class of SuberBase where data is serialized Serder Subclass instance
    given by .klas
    Automatically serializes and deserializes using .klas Serder methods
    """

    def __init__(self, *pa,
                 klas: Type[serdering.Serder] = serdering.SerderKERI,
                 **kwa):
        """
        Inherited Parameters:
            db (dbing.LMDBer): base db
            subkey (str):  LMDB sub database key
            dupsort (bool): True means enable duplicates at each key
                               False (default) means do not enable duplicates at
                               each key
            sep (str): separator to convert keys iterator to key bytes for db key
                       default is self.Sep == '.'
            verify (bool): True means reverify when ._des from db when applicable
                           False means do not reverify. Default False

        Overridden Parameters:
            klas (Type[serdering.Serder]): Class reference to subclass of Serder
        """
        super(SerderSuberBase, self).__init__(*pa, **kwa)
        self.klas = klas


    def _ser(self, val: serdering.Serder):
        """
        Serialize value to bytes to store in db
        Parameters:
            val (serdering.Serder): instance Serder subclass like SerderKERI
        """
        return val.raw


    def _des(self, val: (str | memoryview | bytes)):
        """
        Deserialize val to str
        Parameters:
            val (Union[str, memoryview, bytes]): convertable to coring.matter
        """
        if isinstance(val, memoryview):  # memoryview is always bytes
            val = bytes(val)  # convert to bytes
        elif hasattr(val, "encode"):  # str
            val = val.encode()  # convert to bytes
        return self.klas(raw=val, verify=self.verify)


class SerderSuber(SerderSuberBase, Suber):
    """
    Sub class of SerderSuberBase, Suber where data is serialized Serder Subclass
    instance given by .klas
    Automatically serializes and deserializes using .klas Serder methods
    """

    def __init__(self, *pa, **kwa):
        """
        Inherited Parameters:
            db (dbing.LMDBer): base db
            subkey (str):  LMDB sub database key
            dupsort (bool): True means enable duplicates at each key
                               False (default) means do not enable duplicates at
                               each key
            sep (str): separator to convert keys iterator to key bytes for db key
                       default is self.Sep == '.'
            verify (bool): True means reverify when ._des from db when applicable
                           False means do not reverify. Default False
            klas (Type[serdering.Serder]): Class reference to subclass of Serder
        """
        super(SerderSuber, self).__init__(*pa, **kwa)


class SerderIoSetSuber(SerderSuberBase, IoSetSuber):
    """
    Sub class of SerderSuberBase and IoSetSuber that allows multiple Serder
    instances to be stored at the same db key in insertion order.
    Example use case would be an escrow where the key is a sequence number
    based index (such as snKey).

    Sub class of SerderSuberBase where data is serialized Serder Subclass instance
    given by .klas
    Automatically serializes and deserializes using .klas Serder methods

    Extends IoSetSuber so that all IoSetSuber methods now work with Serder
    subclass for each val.

    IoSetSuber stores at each effective key a set of distinct values that
    share that same effective key where each member of the set is retrieved in
    insertion order (dupsort==False)
    The methods allows an Iterable (set valued) of Iterables of separation subclass
    instances to be stored at a given effective key in insertion order.

    Actual keys include a hidden ordinal key suffix that tracks insertion order.
    The suffix is appended and stripped transparently from the keys. The set of
    items with duplicate effective keys are retrieved in insertion order when
    iterating or as a list of the set elements. The actual iokey for any item
    includes the ordinal suffix.

     Attributes:
        db (dbing.LMDBer): base LMDB db
        sdb (lmdb._Database): instance of lmdb named sub db for this Suber
        sep (str): separator for combining keys tuple of strs into key bytes
        klas (Iterable): of Class references to subclasses of CESR compatible
                , each of to Type[coring.Matter etc]

    """

    def __init__(self, *pa, **kwa):
        """
        Inherited Parameters:
            db (dbing.LMDBer): base db
            subkey (str):  LMDB sub database key
            dupsort (bool): True means enable duplicates at each key
                               False (default) means do not enable duplicates at
                               each key
            sep (str): separator to convert keys iterator to key bytes for db key
                       default is self.Sep == '.'
            verify (bool): True means reverify when ._des from db when applicable
                           False means do not reverify. Default False
            klas (Type[serdering.Serder]): Class reference to subclass of Serder

        """
        super(SerderIoSetSuber, self).__init__(*pa, **kwa)




class SchemerSuber(SerderSuberBase, Suber):
    """
    Sub class of SerderSuberBase and Suber where data is serialized Schemer instance
    Schemer ser/des is ducktype of Serder using .raw
    Automatically serializes and deserializes using Schemer methods
    """

    def __init__(self, *pa,
                 klas: Type[ scheming.Schemer] = scheming.Schemer,
                 **kwa):
        """
        Inherited Parameters:
            db (dbing.LMDBer): base db
            subkey (str):  LMDB sub database key
            dupsort (bool): True means enable duplicates at each key
                               False (default) means do not enable duplicates at
                               each key
            sep (str): separator to convert keys iterator to key bytes for db key
                       default is self.Sep == '.'
            verify (bool): True means reverify when ._des from db when applicable
                           False means do not reverify. Default False
            klas (Type[scheming.Schemer]): Class reference to ducktyped subclass
                of Serder

        Overridden Parameters:
            klas (Type[scheming.Schemer]): Class reference to ducktyped subclass
                of Serder  intercepts passed in klas and forces it to Schemer
        """
        if not issubclass(klas, scheming.Schemer):
            raise TypeError(f"Invalid {klas=}, not subclass of {scheming.Schemer}.")
        super(SchemerSuber, self).__init__(*pa, klas=klas, **kwa)


class DupSuber(SuberBase):
    """
    Sub DB of LMDBer. Subclass of SuberBase that supports multiple entries at
    each key (duplicates) with dupsort==True

    Do not use if  serialized value is greater than 511 bytes.
    This is a limitation of dupsort==True sub dbs in LMDB
    """

    def __init__(self, db: Type[dbing.LMDBer], *,
                       subkey: str='docs.',
                       dupsort: bool=True,
                       **kwa):
        """
        Parameters:
            db (dbing.LMDBer): base db
            subkey (str):  LMDB sub database key
        """
        super(DupSuber, self).__init__(db=db, subkey=subkey, dupsort=True, **kwa)


    def put(self, keys: str | bytes | memoryview | Iterable,
                  vals: str | bytes | memoryview | Iterable):
        """
        Puts all vals at key made from keys. Does not overwrite. Adds to existing
        dup values at key if any. Duplicate means another entry at the same key
        but the entry is still a unique value. Duplicates are inserted in
        lexocographic order not insertion order. Lmdb does not insert a duplicate
        unless it is a unique value for that key.

        Parameters:
            keys (str | bytes | memoryview | Iterable): of key strs to be
                combined in order to form key
            vals (str | bytes | memoryview | Iterable): str or bytes of each
                value to be written at key

        Returns:
            result (bool): True If successful, False otherwise.

        Apparently always returns True (how .put works with dupsort=True)

        """
        if not nonStringIterable(vals):  # not iterable
            vals = (vals, )  # make iterable
        return (self.db.putVals(db=self.sdb,
                                key=self._tokey(keys),
                                vals=[self._ser(val) for val in vals]))


    def add(self, keys: str | bytes | memoryview | Iterable,
                  val: str | bytes | memoryview ):
        """
        Add val to vals at key made from keys. Does not overwrite. Adds to existing
        dup values at key if any. Duplicate means another entry at the same key
        but the entry is still a unique value. Duplicates are inserted in
        lexocographic order not insertion order. Lmdb does not insert a duplicate
        unless it is a unique value for that key.

        Parameters:
            keys (str | bytes | memoryview | Iterable): of key strs to be combined in order to form key
            val (str | bytes | memoryview): value

        Returns:
            result (bool): True means unique value among duplications,
                              False means duplicte of same value already exists.

        """
        return (self.db.addVal(db=self.sdb,
                               key=self._tokey(keys),
                               val=self._ser(val)))


    def pin(self, keys: str | bytes | memoryview | Iterable,
                  vals: str | bytes | memoryview | Iterable):
        """
        Pins (sets) vals at key made from keys. Overwrites. Removes all
        pre-existing dup vals and replaces them with vals

        Parameters:
            keys (str | bytes | memoryview | Iterable): of key strs to be
                combined in order to form key
            vals (str | bytes | memoryview | Iterable): str or bytes values

        Returns:
            result (bool): True If successful, False otherwise.

        """
        key = self._tokey(keys)
        self.db.delVals(db=self.sdb, key=key)  # delete all values
        if not nonStringIterable(vals):  # not iterable
            vals = (vals, )  # make iterable
        return (self.db.putVals(db=self.sdb,
                                key=key,
                                vals=[self._ser(val) for val in vals]))



    def get(self, keys: str | bytes | memoryview | Iterable):
        """
        Gets dup vals list at key made from keys

        Parameters:
            keys (str | bytes | memoryview | Iterable): of key strs to be
                combined in order to form key

        Returns:
            vals (list):  each item in list is str
                          empty list if no entry at keys

        """
        return [self._des(val) for val in
                        self.db.getValsIter(db=self.sdb, key=self._tokey(keys))]


    def getLast(self, keys: str | bytes | memoryview | Iterable):
        """
        Gets last dup val at key made from keys

        Parameters:
            keys (tuple): of key strs to be combined in order to form key

        Returns:
            val (str):  value else None if no value at key

        """
        val = self.db.getValLast(db=self.sdb, key=self._tokey(keys))
        return self._des(val) if val is not None else val


    def getIter(self, keys: str | bytes | memoryview | Iterable):
        """
        Gets dup vals iterator at key made from keys

        Duplicates are retrieved in lexocographic order not insertion order.

        Parameters:
            keys (str | bytes | memoryview | Iterable): of key strs to be
                combined in order to form key

        Returns:
            iterator:  vals each of str. Raises StopIteration when done

        """
        for val in self.db.getValsIter(db=self.sdb, key=self._tokey(keys)):
            yield self._des(val)


    def cnt(self, keys: str | bytes | memoryview | Iterable):
        """
        Return count of dup values at key made from keys, zero otherwise

        Parameters:
            keys (str | bytes | memoryview | Iterable): of key strs to be
                combined in order to form key
        """
        return (self.db.cntVals(db=self.sdb, key=self._tokey(keys)))


    def rem(self, keys: str | bytes | memoryview | Iterable,
            val: str | bytes | memoryview = b''):
        """
        Removes entry at keys

        Parameters:
            keys (tuple): of key strs to be combined in order to form key
            val (Union[str, bytes]):  instance of dup val at key to delete
                              if val is None then remove all values at key

        Returns:
           result (bool): True if key exists so delete successful. False otherwise

        """
        if val:
            return (self.db.delVals(db=self.sdb,
                                    key=self._tokey(keys),
                                    val=self._ser(val)))
        else:
            return (self.db.delVals(db=self.sdb,
                                    key=self._tokey(keys)))



class CesrDupSuber(CesrSuberBase, DupSuber):
    """
    Sub class of DupSuber whose values are CESR ducktypes of Matter subclasses.
    serialized to and deserialied from val instance .qb64b property
    which is a fully qualified serialization.
    Automatically serializes and deserializes from qb64b to/from Matter ducktyped
    instances
    DupSuber supports multiple entries at each key (duplicates) with dupsort==True

    Do not use if  serialized value is greater than 511 bytes.
    This is a limitation of dupsort==True sub dbs in LMDB
    """
    def __init__(self, *pa, **kwa):
        """
        Inherited Parameters:

        """
        super(CesrDupSuber, self).__init__(*pa, **kwa)


class IoDupSuber(DupSuber):
    """
    Sub class of DupSuber that supports Insertion Ordering (IoDup) of duplicates
    By automagically prepending and stripping  ordinal proem to/from each
    duplicate value at a given key.

    IoDupSuber supports  insertion ordered multiple entries at each key
    (duplicates) with dupsort==True

    Do not use if  serialized length key + proem + value, is greater than 511 bytes.
    This is a limitation of dupsort==True sub dbs in LMDB

    IoDupSuber may be more performant then IoSetSuber for values that are indices
    to other sub dbs that fit the size constraint because LMDB support for
    duplicates is more space efficient and code performant.

    Duplicates at a given key preserve insertion order of duplicate.
    Because lmdb is lexocographic an insertion ordering proem is prepended to
    all values that makes lexocographic order that same as insertion order.

    Duplicates are ordered as a pair of key plus value so prepending proem
    to each value changes duplicate ordering. Proem is 33 characters long.
    With 32 character hex string followed by '.' for essentiall unlimited
    number of values which will be limited by memory.

    With prepended proem ordinal must explicity check for duplicate values
    before insertion. Uses a python set for the duplicate inclusion test.
    Set inclusion scales with O(1) whereas list inclusion scales with O(n).
    """
    def __init__(self, *pa, **kwa):
        """
        Inherited Parameters:

        """
        super(IoDupSuber, self).__init__(*pa, **kwa)


    def put(self, keys: str | bytes | memoryview | Iterable,
                  vals: str | bytes | memoryview | Iterable):
        """
        Puts all vals idempotently at key made from keys in insertion order using
        hidden ordinal proem. Idempotently means do not put any val in vals that is
        already in dup vals at key. Does not overwrite.

        Parameters:
            keys (Iterable): of key strs to be combined in order to form key
            vals (Iterable): of str serializations

        Returns:
            result (bool): True If successful, False otherwise.

        """
        if not nonStringIterable(vals):  # not iterable
            vals = (vals, )  # make iterable
        return (self.db.putIoDupVals(db=self.sdb,
                                     key=self._tokey(keys),
                                     vals=[self._ser(val) for val in vals]))


    def add(self, keys: str | bytes | memoryview | Iterable,
                  val: str | bytes | memoryview):
        """
        Add val idempotently  at key made from keys in insertion order using hidden
        ordinal proem. Idempotently means do not add val that is already in
        dup vals at key. Does not overwrite.

        Parameters:
            keys (Iterable): of key strs to be combined in order to form key
            val (str | bytes | memoryview): serialization

        Returns:
            result (bool): True means unique value added among duplications,
                            False means duplicate of same value already exists.

        """
        return (self.db.addIoDupVal(db=self.sdb,
                                    key=self._tokey(keys),
                                    val=self._ser(val)))


    def pin(self, keys: str | bytes | memoryview | Iterable,
            vals: str | bytes | memoryview | Iterable):
        """
        Pins (sets) vals at key made from keys in insertion order using hidden
        ordinal proem. Overwrites. Removes all pre-existing vals that share
        same keys and replaces them with vals

        Parameters:
            keys (Iterable): of key strs to be combined in order to form key
            vals (Iterable): str serializations

        Returns:
            result (bool): True If successful, False otherwise.

        """
        key = self._tokey(keys)
        self.db.delIoDupVals(db=self.sdb, key=key)  # delete all values
        if not nonStringIterable(vals):  # not iterable
            vals = (vals, )  # make iterable
        return self.db.putIoDupVals(db=self.sdb,
                                     key=key,
                                     vals=[self._ser(val) for val in vals])


    def get(self, keys: str | bytes | memoryview | Iterable):
        """
        Gets vals dup list in insertion order using key made from keys and
        hidden ordinal proem on dups.

        Parameters:
            keys (Iterable): of key strs to be combined in order to form key

        Returns:
            vals (Iterable):  each item in list is str
                          empty list if no entry at keys

        """
        return ([self._des(val) for val in
                    self.db.getIoDupVals(db=self.sdb, key=self._tokey(keys))])


    def getIter(self, keys: str | bytes | memoryview | Iterable):
        """
        Gets vals dup iterator in insertion order using key made from keys and
        hidden ordinal proem on dups.
        All vals in dups that share same key are retrieved in insertion order.

        Parameters:
            keys (str | bytes | memoryview | Iterable): of key parts

        Returns:
            vals (Iterator):  str values. Raises StopIteration when done

        """
        for val in self.db.getIoDupValsIter(db=self.sdb,
                                            key=self._tokey(keys)):
            yield self._des(val)


    def getLast(self, keys: str | bytes | memoryview | Iterable):
        """
        Gets last val inserted at key made from keys in insertion order using
        hidden ordinal proem.

        Parameters:
            keys (Iterable): of key strs to be combined in order to form key

        Returns:
            val (str):  value str, None if no entry at keys

        """
        val = self.db.getIoDupValLast(db=self.sdb, key=self._tokey(keys))
        return (self._des(val) if val is not None else val)


    def rem(self, keys: str | bytes | memoryview | Iterable,
                   val: str | bytes | memoryview = ''):
        """
        Removes entry at key made from keys and dup val that matches val if any,
        notwithstanding hidden ordinal proem. Otherwise deletes all dup values
        at key if any.

        Parameters:
            keys (str | bytes | memoryview | Iterable): of key parts to be
                combined in order to form key
            val (str):  value at key to delete. Subclass ._ser method may
                        accept different value types
                        if val is empty then remove all values at key

        Returns:
           result (bool): True if key with dup val exists so rem successful.
                           False otherwise

        """
        if val:
            return self.db.delIoDupVal(db=self.sdb,
                                       key=self._tokey(keys),
                                       val=self._ser(val))
        else:
            return self.db.delIoDupVals(db=self.sdb, key=self._tokey(keys))


    def cnt(self, keys: str | bytes | memoryview | Iterable):
        """
        Return count of dup values at key made from keys with hidden ordinal
        proem. Zero otherwise

        Parameters:
            keys (str | bytes | memoryview | Iterable): of key parts to be
                combined in order to form key
        """
        return (self.db.cntIoDupVals(db=self.sdb, key=self._tokey(keys)))


    def getItemIter(self, keys: str | bytes | memoryview | Iterable = "",
                    *, topive=False):
        """
        Return iterator over all the items including dup items for all keys
        in top branch defined by keys where keys may be truncation of full branch.

        Returns:
            items (Iterator): of (key, val) tuples over the all the items in
            subdb whose key startswith key made from keys and val has its hidden
            dup ordinal proem removed.
            Keys may be keyspace prefix in order to return branches of key space.
            When keys is empty then returns all items in subdb.

        Parameters:
            keys (str | bytes | memoryview | Iterable): key or key parts that
                may be a truncation of a full keys tuple in  in order to address
                all the items from multiple branches of the key space.
                If keys is empty then gets all items in database.
                Either append "" to end of keys Iterable to ensure get properly
                separated top branch key or use top=True.
                In Python str.startswith('') always returns True so if branch
                key is empty string it matches all keys in db with startswith.

            topive (bool): True means treat as partial key tuple from top branch of
                key space given by partial keys. Resultant key ends in .sep
                character.
                False means treat as full branch in key space. Resultant key
                does not end in .sep character.
                When last item in keys is empty str then will treat as
                partial ending in sep regardless of top value

        """

        for key, val in self.db.getTopIoDupItemIter(db=self.sdb,
                                         top=self._tokey(keys, topive=topive)):
            yield (self._tokeys(key), self._des(val))


class OnIoDupSuber(OnSuberBase, IoDupSuber):
    """
    Sub class of IoDupSuber and OnSuberBase that supports Insertion Ordering
    (IoDup) of duplicates but where the trailing part of the key space is
    a serialized monotonically increasing ordinal number. This is useful for
    escrows of key events etc where duplicates of likely events are maintained
    in insertion order.
    Insertion order is maintained by automagically prepending and stripping an
    ordinal ordering proem to/from each duplicate value at a given key.

    OnIoDupSuber adds the convenience methods from OnSuberBase to IoDupSuber for
    those cases where the keyspace has a trailing ordinal part.

    There are two ordinals, one in the key space and a hidden one in the duplicate
    data value space.

    OnIoDupSuber supports  insertion ordered multiple entries at each key
    (duplicates) with dupsort==True

    Do not use if  serialized length key + proem + value, is greater than 511 bytes.
    This is a limitation of dupsort==True sub dbs in LMDB

    OnIoDupSuber may be more performant then IoSetSuber for values that are indices
    to other sub dbs that fit the size constraint because LMDB support for
    duplicates is more space efficient and code performant.

    Duplicates at a given key preserve insertion order of duplicate.
    Because lmdb is lexocographic an insertion ordering proem is prepended to
    all values that makes lexocographic order that same as insertion order.

    Duplicates are ordered as a pair of key plus value so prepending proem
    to each value changes duplicate ordering. Proem is 33 characters long.
    With 32 character hex string followed by '.' for essentiall unlimited
    number of values which will be limited by memory.

    With prepended proem ordinal must explicity check for duplicate values
    before insertion. Uses a python set for the duplicate inclusion test.
    Set inclusion scales with O(1) whereas list inclusion scales with O(n).
    """
    def __init__(self, *pa, **kwa):
        """
        Inherited Parameters:

        """
        super(OnIoDupSuber, self).__init__(*pa, **kwa)


    def addOn(self, keys: str | bytes | memoryview | Iterable, on: int=0,
                  val: str | bytes | memoryview = ''):
        """
        Add val idempotently  at key made from keys in insertion order using hidden
        ordinal proem. Idempotently means do not add val that is already in
        dup vals at key. Does not overwrite.

        Parameters:
            keys (str | bytes | memoryview | Iterable): top keys as prefix to be
                combined with serialized on suffix and sep to form onkey
            on (int): ordinal number used with onKey(pre,on) to form onkey.
            val (str | bytes | memoryview): serialization

        Returns:
            result (bool): True means unique value added among duplications,
                            False means duplicate of same value already exists.

        """
        return (self.db.addOnIoDupVal(db=self.sdb,
                                    key=self._tokey(keys),
                                    on=on,
                                    val=self._ser(val),
                                    sep=self.sep.encode()))



    def appendOn(self, keys: str | bytes | memoryview,
                       val: str | bytes | memoryview):
        """
        Returns:
            on (int): ordinal number of newly appended val

        Parameters:
            keys (str | bytes | memoryview | Iterable): top keys as prefix to be
                combined with serialized on suffix and sep to form key
            val (str | bytes | memoryview): serialization
        """
        return (self.db.appendOnIoDupVal(db=self.sdb,
                                       key=self._tokey(keys),
                                       val=self._ser(val),
                                       sep=self.sep.encode()))


    def getOn(self, keys: str | bytes | memoryview | Iterable, on: int = 0):
        """
        Gets dup vals list at key made from keys

        Parameters:
            keys (str | bytes | memoryview | Iterable): of key strs to be
                combined in order to form key
            on (int): ordinal number used with onKey(pre,on) to form key.

        Returns:
            vals (list):  each item in list is str
                          empty list if no entry at keys

        """
        return [self._des(val) for val in
                        self.db.getOnIoDupValIter(db=self.sdb,
                                                  key=self._tokey(keys),
                                                  on=on,
                                                  sep=self.sep.encode())]


    def remOn(self, keys: str | bytes | memoryview | Iterable, on: int=0,
                   val: str | bytes | memoryview = ''):
        """
        Removes entry at key made from keys and dup val that matches val if any,
        notwithstanding hidden ordinal proem. Otherwise deletes all dup values
        at key if any.

        Parameters:
            keys (str | bytes | memoryview | iterator): keys as prefix to be
                combined with serialized on suffix and sep to form onkey

            on (int): ordinal number used with onKey(pre,on) to form key.
            val (str):  value at key to delete. Subclass ._ser method may
                        accept different value types
                        if val is empty then remove all values at key

        Returns:
           result (bool): True if onkey with dup val exists so rem successful.
                           False otherwise

        """
        if val:
            return self.db.delOnIoDupVal(db=self.sdb,
                                       key=self._tokey(keys),
                                       on=on,
                                       val=self._ser(val),
                                       sep=self.sep.encode())
        else:
            return self.db.delOnIoDupVals(db=self.sdb,
                                          key=self._tokey(keys),
                                          on=on,
                                          sep=self.sep.encode())



    def getOnIter(self, keys: str|bytes|memoryview|Iterable = "", on: int=0):
        """
        Returns
            val (Iterator[bytes]):  deserialized val of of each
                onkey

        Parameters:
            keys (str | bytes | memoryview | iterator): keys as prefix to be
                combined with serialized on suffix and sep to form onkey
                When keys is empty then retrieves whole database including duplicates
            on (int): ordinal number used with onKey(pre,on) to form key.
            sep (bytes): separator character for split
        """
        for val in (self.db.getOnIoDupValIter(db=self.sdb,
                        key=self._tokey(keys), on=on, sep=self.sep.encode())):
            yield (self._des(val))


    def getOnItemIter(self, keys: str|bytes|memoryview|Iterable = "", on: int=0):
        """
        Returns:
            items (Iterator[(top keys, on, val)]): triples of (onkeys, on int,
                  deserialized val)

        Parameters:
            keys (str | bytes | memoryview | iterator): keys as prefix to be
                combined with serialized on suffix and sep to form onkey
                When keys is empty then retrieves whole database including duplicates
            on (int): ordinal number used with onKey(pre,on) to form key.
            sep (bytes): separator character for split
        """
        for keys, on, val in (self.db.getOnIoDupItemIter(db=self.sdb,
                        key=self._tokey(keys), on=on, sep=self.sep.encode())):
            yield (self._tokeys(keys), on, self._des(val))


    def getOnLastIter(self, keys: str|bytes|memoryview|Iterable = "", on: int=0):
        """
        Returns
            last (Iterator[bytes]):  deserialized last duplicate val of of each
                onkey

        Parameters:
            keys (str | bytes | memoryview | iterator): top keys as prefix to be
                combined with serialized on suffix and sep to form key
                When keys is empty then retrieves whole database including duplicates
            on (int): ordinal number used with onKey(pre,on) to form key.
            sep (bytes): separator character for split
        """
        for val in (self.db.getOnIoDupLastValIter(db=self.sdb,
                        key=self._tokey(keys), on=on, sep=self.sep.encode())):
            yield (self._des(val))



    def getOnLastItemIter(self, keys: str|bytes|memoryview|Iterable = "", on: int=0):
        """
        Returns
            items (Iterator[(top keys, on, val)]): triples of (keys, on int,
                  deserialized val) last duplicate item as each onkey where onkey
                  is the key+serialized on

        Parameters:
            keys (str | bytes | memoryview | iterator): keys as prefix to be
                combined with serialized on suffix and sep to form key
                When keys is empty then retrieves whole database including duplicates
            on (int): ordinal number used with onKey(pre,on) to form key.
            sep (bytes): separator character for split
        """
        for keys, on, val in (self.db.getOnIoDupLastItemIter(db=self.sdb,
                        key=self._tokey(keys), on=on, sep=self.sep.encode())):
            yield (self._tokeys(keys), on, self._des(val))


    def getOnBackIter(self, keys: str|bytes|memoryview|Iterable = "", on: int=0):
        """
        Returns
            val (Iterator[bytes]):  deserialized val of of each
                onkey in reverse order

        Parameters:
            keys (str | bytes | memoryview | iterator): keys as prefix to be
                combined with serialized on suffix and sep to form onkey
                When keys is empty then retrieves whole database including duplicates
            on (int): ordinal number used with onKey(pre,on) to form key.
            sep (bytes): separator character for split
        """
        for val in (self.db.getOnIoDupValBackIter(db=self.sdb,
                        key=self._tokey(keys), on=on, sep=self.sep.encode())):
            yield (self._des(val))


    def getOnItemBackIter(self, keys: str|bytes|memoryview|Iterable = "", on: int=0):
        """
        Returns:
            items (Iterator[(top keys, on, val)]): triples of (onkeys, on int,
                  deserialized val) in reverse order

        Parameters:
            keys (str | bytes | memoryview | iterator): keys as prefix to be
                combined with serialized on suffix and sep to form onkey
                When keys is empty then retrieves whole database including duplicates
            on (int): ordinal number used with onKey(pre,on) to form key.
            sep (bytes): separator character for split
        """
        for keys, on, val in (self.db.getOnIoDupItemBackIter(db=self.sdb,
                        key=self._tokey(keys), on=on, sep=self.sep.encode())):
            yield (self._tokeys(keys), on, self._des(val))



# -*- encoding: utf-8 -*-
"""
KERI
keri.core Package
"""

# Constants etc
from .coring import (Tiers, )

# Matter class and its subclasses
from .coring import (Matter, MtrDex, Number, NumDex, Dater, Texter,
                    Bexter, Pather, Verfer, Cigar, Diger, DigDex,
                    Prefixer, PreDex, )

from .coring import Tholder
from .indexing import Indexer, Siger, IdrDex, IdxSigDex
from .signing import Signer, Salter, Cipher, CiXDex, Encrypter, Decrypter
from .counting import Counter, Codens, CtrDex_2_0
from .streaming import Streamer



# -*- encoding: utf-8 -*-
"""
KERI
keri.app.watching module

"""
import random
from collections import namedtuple
from dataclasses import dataclass

from hio.base import doing
from hio.help import decking

from keri import help

logger = help.ogler.getLogger()

Stateage = namedtuple("Stateage", 'even ahead behind duplicitous')

States = Stateage(even="even", ahead="ahead", behind="behind", duplicitous="duplicitous")


@dataclass
class DiffState:
    """ Difference between a remote KeyStateRecord and local for the same AID.

    Uses Stateage to represent whether the remote KSR is even, ahead, behind or duplicitous

    """
    pre: str  # The AID of the KSR
    wit: str  # The entity reporting the KSR (non-local)
    state: Stateage  # The state of the remote KSR relative to local
    sn: int  # The sequence number of the remote KSR
    dig: str  # The digest of the latest event of the remote KSR


class Adjudicator:
    """ The Adjudicator of Key State

    This class performs key state adjudication by checking any key state reported by the watcher set for a given
    watched AID and compares the reported values against the local key state for the watched AID and the key state
    of all other responding watchers.  It uses a per-adjudication threshold to determine what is acceptable duplicity
    for each adjudication.

    Cues are sent out for each round of adjudication with the following kins:

       keyStateConsistent - Key state of all queries watchers is consistent with local key state
       keyStateLagging - Key state from some watchers is behind local key state and other watchers
       keyStateUpdate - A threshold satisfying number of watchers report new key state for watched AID
       keyStateDuplicitous - Duplicity has been detected on some set of watchers (provided in the cue)

    Consumers of the Adjudicator's cues are safe to retrieve new key state from one of the Watchers listed in the
    cue of `keyStateUpdated` is received.  All other kins require controller intervention and should be bubbled up.

    """

    def __init__(self, hby, hab, msgs=None, cues=None):
        """ Create instance of Adjudicator for adjudicating key state

        Parameters:
            hby (Habery): database and Habitat environment
            hab (Hab): identifier database environment
            msgs (Deck): incoming requests to adjudicate key state
            cues (Deck): outgoing responses to adjudication of key state

        """
        self.hby = hby
        self.hab = hab
        self.msgs = msgs if msgs is not None else decking.Deck()
        self.cues = cues if cues is not None else decking.Deck()

    def performAdjudications(self):
        """ Process loop of existing messages requesting key state adjudication """
        while self.msgs:
            msg = self.msgs.pull()

            watched = msg["oid"]
            toad = msg["toad"] if "toad" in msg else None

            self.adjudicate(watched, toad)

    def adjudicate(self, watched, toad=None):
        """ Perform key state adjudication against the `watched` AID and provided threshold

        If `toad` is not provided, the full set of watchers must come to consensus before `keyStateUpdate`
        will be reported.

        Parameters:
            watched (str): qb64 AID to adjudicate for key state duplicity
            toad (int): threshold of acceptable duplicity amongst available watchers


        """
        watchers = set()
        for (cid, aid, oid), observed in self.hab.db.obvs.getItemIter(keys=(self.hab.pre,)):
            if observed.enabled and oid == watched:
                watchers.add(aid)

        toad = int(toad) if toad else len(watchers)
        if toad > len(watchers):
            raise ValueError(f"Threshold of {toad} is greater than number watchers {len(watchers)}")

        states = []
        mystate = self.hab.kevers[watched].state()
        for watcher in watchers:
            saider = self.hab.db.knas.get(keys=(watched, watcher))
            if saider is None:
                logger.info(f"No key state from watcher {watcher} for {watched}")
                continue

            ksn = self.hab.db.ksns.get(keys=(saider.qb64,))
            states.append(diffState(watcher, mystate, ksn))

        dups = [state for state in states if state.state == States.duplicitous]
        ahds = [state for state in states if state.state == States.ahead]
        bhds = [state for state in states if state.state == States.behind]

        if len(dups) > 0:
            cue = dict(kin="keyStateDuplicitous", cid=self.hab.pre, oid=watched, wids=watchers, dups=dups)
            self.cues.append(cue)

            logger.error(f"Duplicity detected for AID {watched}, local key state remains intact.")
            for state in dups:
                logger.error(f"\tWatcher {state.wit} at seq No. {state.sn} with digest: {state.dig}")

        elif len(ahds) > 0:
            # Only group habs can be behind their watchers
            # First check for duplicity among the watchers that are ahead (possible only if toad is below
            # super majority)
            digs = set([state.dig for state in ahds])
            if len(digs) > 1:  # Duplicity across watcher sets
                cue = dict(kin="keyStateDuplicitous", cid=self.hab.pre, oid=watched, wids=watchers, dups=ahds)
                self.cues.append(cue)

                logger.error(f"There are multiple duplicitous events on watcher for {watched}")
                for state in ahds:
                    logger.error(f"\tWatcher {state.wit} at seq No. {state.sn} with digest: {state.dig}")

            elif len(ahds) >= toad:  # all witnesses that are ahead agree on the event
                logger.info(f"Threshold ({toad}) satisfying number of watchers ({len(ahds)}) are ahead")
                for state in ahds:
                    logger.info(f"\tWatcher {state.wit} at Seq No. {state.sn} with digest: {state.dig}")

                state = random.choice(ahds)
                cue = dict(kin="keyStateUpdate", cid=self.hab.pre, oid=watched, wids=watchers, sn=state.sn, aheads=ahds)
                self.cues.append(cue)

        elif len(bhds) > 0:
            cue = dict(kin="keyStateLagging", cid=self.hab.pre, oid=watched, wids=watchers, behind=bhds)
            self.cues.append(cue)

            logger.info("The following watchers are behind the local KEL:")
            for state in bhds:
                logger.info(f"\tWatcher {state.wit} at seq No. {state.sn} with digest: {state.dig}")

            logger.info(f"Recommend the checking those watchers for access to {watched} witnesses")

        else:
            cue = dict(kin="keyStateConsistent", cid=self.hab.pre, oid=watched, wids=watchers, states=states)
            self.cues.append(cue)
            logger.info(f"Local key state is consistent with the {len(states)} (out of "
                        f"{len(watchers)} total) watchers that responded")


class AdjudicationDoer(doing.Doer):
    """ Doer class responsible for process adjudication requests in an Adjudicator's msgs """

    def __init__(self, adjudicator):
        """ Create instance of Doer for performing key state adjudications """
        self.adjudicator = adjudicator
        super(AdjudicationDoer, self).__init__()

    def recur(self, tyme):
        """ Perform one pass over all adjudication requests

        Parameters:
            tyme (float): relative cycle time

        Returns:

        """
        self.adjudicator.performAdjudications()


def diffState(wit, preksn, witksn):
    """ Return a record of the differences between the states provided by `wit` and local state

    Parameters:
        wit (str): qb64 AID of entity reporting key state
        preksn (KeyStateRecord): Local key state of AID
        witksn (KeyStateRecord): Key state of AID as provided by `wit`

    Returns:
        state (WitnessState): record indicating the differenced between the two provided KSN records

    """
    mypre = preksn.i
    pre = witksn.i
    mysn = int(preksn.s, 16)
    mydig = preksn.d
    sn = int(witksn.s, 16)
    dig = witksn.d

    if pre != mypre:
        raise ValueError(f"can't compare key states from different AIDs {mypre}/{pre}")

    # At the same sequence number, check the DIGs
    if mysn == sn:
        if mydig == dig:
            state = States.even
        else:
            state = States.duplicitous

    # This witness is behind and will need to be caught up.
    elif mysn > sn:
        state = States.behind

    # mysn < witstate.sn - We are behind this witness (multisig or restore situation).
    # Must ensure that controller approves this event or a recovery rotation is needed
    else:
        state = States.ahead

    return DiffState(pre, wit, state, sn, dig)



# -*- encoding: utf-8 -*-
"""
keri.kli.commands module

"""
import argparse

from hio.base import doing

from keri.core import serdering
from ..common import existing
from ... import habbing, connecting, forwarding

parser = argparse.ArgumentParser(description='Send an rpy /introduce message to recipient with OOBI')
parser.set_defaults(handler=lambda args: introduce(args))
parser.add_argument('--name', '-n', help='keystore name and file location of KERI keystore', required=True)
parser.add_argument('--base', '-b', help='additional optional prefix to file location of KERI keystore',
                    required=False, default="")
parser.add_argument('--alias', '-a', help='human readable alias for the new identifier prefix', required=True)
parser.add_argument('--passcode', '-p', help='21 character encryption passcode for keystore (is not saved)',
                    dest="bran", default=None)  # passcode => bran
parser.add_argument('--recipient', help='alias or contact to send the introduction to', required=True)
parser.add_argument('--introducee', help='alias or contact of the OOBI to send', required=True)
parser.add_argument("--role", "-r", help="role of oobi to send", required=True)


def introduce(args):
    """
    Performs a rotation of the identifier of the environment represented by the provided name parameter

        args (parseargs):  Command line argument

    """
    name = args.name
    alias = args.alias
    base = args.base
    bran = args.bran
    introducee = args.introducee
    recipient = args.recipient
    role = args.role

    ixnDoer = IntroduceDoer(name=name, base=base, alias=alias, bran=bran, introducee=introducee, recipient=recipient,
                            role=role,)

    return [ixnDoer]


class IntroduceDoer(doing.DoDoer):
    """

       DoDoer that launches Doers needed to introduce one controller of an AID to another

    """

    def __init__(self, name, base, bran, alias, introducee, recipient, role):
        """
        Returns DoDoer with all registered Doers needed to perform an introduction.

        Parameters:
            name is human readable str of identifier
       """

        self.alias = alias
        self.introducee = introducee
        self.recipient = recipient
        self.role = role

        self.hby = existing.setupHby(name=name, base=base, bran=bran)
        self.hbyDoer = habbing.HaberyDoer(habery=self.hby)  # setup doer
        self.org = connecting.Organizer(hby=self.hby)
        self.postman = forwarding.Poster(hby=self.hby)
        doers = [self.hbyDoer, self.postman, doing.doify(self.introduceDo)]

        super(IntroduceDoer, self).__init__(doers=doers)

    def introduceDo(self, tymth, tock=0.0, **opts):
        """
        Returns:  doifiable Doist compatible generator method
        Usage:
            add result of doify on this method to doers list
        """
        self.wind(tymth)
        self.tock = tock
        _ = (yield self.tock)

        hab = self.hby.habByName(name=self.alias)

        if self.recipient in self.hby.kevers:
            recp = self.recipient
        else:
            recp = self.org.find("alias", self.recipient)
            if len(recp) != 1:
                raise ValueError(f"invalid recipient {self.recipient}")
            recp = recp[0]['id']

        if (ihab := self.hby.habByName(self.introducee)) is not None:
            introducee = ihab.pre
        elif self.introducee in self.hby.habs:
            introducee = self.introducee
        elif self.introducee in self.hby.kevers:
            introducee = self.introducee
        else:
            introducee = None
            results = self.org.find("alias", self.introducee)
            for result in results:
                if result["alias"] == self.introducee:
                    introducee = result['id']
            if not introducee:
                raise ValueError(f"invalid introducee {self.introducee}")

        oobi = None
        for (key,), obr in self.hby.db.roobi.getItemIter():
            if obr.cid == introducee and obr.role == self.role:
                oobi = key

        if oobi is None:
            raise ValueError(f"Unable to find OOBI with role {self.role} for {introducee}")

        msg = hab.reply(route="/introduce", data=dict(
            cid=hab.pre,
            oobi=oobi
        ))

        serder = serdering.SerderKERI(raw=msg)
        atc = msg[serder.size:]
        self.postman.send(src=hab.pre, dest=recp, topic="credential", serder=serder,
                          attachment=atc)

        while not len(self.postman.cues) == 1:
            yield self.tock

        print(f"Introduction with OOBI {oobi} sent for {self.introducee} sent to {self.recipient}")

        toRemove = [self.hbyDoer, self.postman]
        self.remove(toRemove)

        return



# -*- encoding: utf-8 -*-
"""
keri.app.storing module

"""

from hio.base import doing
from hio.help import decking
from ordered_set import OrderedSet as oset

from . import forwarding
from .. import help
from ..core import coring, serdering
from ..core.coring import MtrDex
from ..db import dbing, subing

logger = help.ogler.getLogger()


class Mailboxer(dbing.LMDBer):
    """
    Mailboxer stores exn messages in order and provider iterator access at an index.

    """
    TailDirPath = "keri/mbx"
    AltTailDirPath = ".keri/mbx"
    TempPrefix = "keri_mbx_"

    def __init__(self, name="mbx", headDirPath=None, reopen=True, **kwa):
        """

        Parameters:
            headDirPath:
            perm:
            reopen:
            kwa:

        Mailboxer uses two dbs for mailbox messages these are .tpcs and .msgs.
        The message index is in .tpcs (topics).
            Each .tpcs index key consists of topic.on where topic is bytes
                identifier or prefix/topic for message and on is serialized
                ordinal number to orders the appearance of a topic message.
            Eash .tpcs val is the digest of the message.
        The message itself is stored in .msgs where the key is the msg digest
            and the value is the serialized messag itself.
        Multiple messages can share the same topic but with a different ordinal.

        """
        self.tpcs = None
        self.msgs = None

        super(Mailboxer, self).__init__(name=name, headDirPath=headDirPath, reopen=reopen, **kwa)

    def reopen(self, **kwa):
        """

        :param kwa:
        :return:
        """
        super(Mailboxer, self).reopen(**kwa)
        self.tpcs = subing.OnSuber(db=self, subkey='tpcs.')
        self.msgs = subing.Suber(db=self, subkey='msgs.')  # key states

        return self.env

    def delTopic(self, key, on=0):
        """Removes topic index from .tpcs without deleting message from .msgs

        Returns:
            result (boo): True if full key consisting of key and serialized on
                             exists in database so removed
                          False otherwise (not removed)
        """
        return self.tpcs.remOn(keys=key, on=on)

    def appendToTopic(self, topic, val):
        """Appends val to end of db entries with same topic but with on
        incremented by 1 relative to last preexisting entry at topic.

        Returns:
            on (int): order number int, on, of appended entry.
                      Computes on as next on after last entry.

        Parameters:
            topic (bytes):  topic identifier for message
            val (bytes): msg digest
        """
        return self.tpcs.appendOn(key=topic, val=val)


    def getTopicMsgs(self, topic, fn=0):
        """
        Returns:
            msgs (Iterable[bytes]): belonging to topic indices with same topic but all
                on >= fn i.e. all topic.on beginning with fn

         Parameters:
             topic (Option(bytes|str)): key prefix combined with serialized on
                    to form full actual key.  When key is empty then retrieves
                    whole database.
             fn (int): starting index ordinal number used with onKey(pre,on)
                    to form key at at which to initiate retrieval
        """
        msgs = []
        for keys, on, dig in self.tpcs.getOnItemIter(keys=topic, on=fn):
            if msg := self.msgs.get(keys=dig):
                msgs.append(msg.encode())  # want bytes not str
        return msgs


    def storeMsg(self, topic, msg):
        """
        Add exn event to mailbox topic and on that is 1 greater than last msg
        at topic.

        Returns:
            result (bool): True if msg successfully stored and indexed at topic
                           False otherwise

        Parameters:
            topic (str | bytes):  topic (Option(bytes|str)): key prefix combined
                with serialized on to form full actual key.
            msg (bytes): serialized message

        """
        if hasattr(msg, "encode"):
            msg = msg.encode("utf-8")

        digb = coring.Diger(ser=msg, code=MtrDex.Blake3_256).qb64b
        on = self.tpcs.appendOn(keys=topic, val=digb)
        return self.msgs.pin(keys=digb, val=msg)


    def cloneTopicIter(self, topic, fn=0):
        """
        Returns:
            triple (Iterator[(on, topic, msg): iterator of messages at topic
            beginning with ordinal fn.

        topic (Option(bytes|str)): key prefix combined with serialized on
                    to form full actual key.  When key is empty then retrieves
                    whole database.
             fn (int): starting index ordinal number used with onKey(pre,on)
                    to form key at at which to initiate retrieval



        ToDo looks like misuse of IoSet this should not be IoSet but simply
        Ordinal Numbered db.  since should not be using hidden ion has not
        hidden.

        """
        for keys, on, dig in self.tpcs.getOnItemIter(keys=topic, on=fn):
            if msg := self.msgs.get(keys=dig):
                yield (on, topic, msg.encode("utf-8"))



class Respondant(doing.DoDoer):
    """
    Respondant processes buffer of response messages from inbound 'exn' messages and
    routes them to the appropriate mailbox.  If destination has witnesses, send response to
    one of the (randomly selected) witnesses.  Otherwise store the response in the recipients
    mailbox locally.

    """

    def __init__(self, hby, reps=None, cues=None, mbx=None, aids=None, **kwa):
        """
        Creates Respondant that uses local environment to find the destination KEL and stores
        peer to peer messages in mbx, the mailboxer

        Parameters:
            hab (Habitat):  local environment
            mbx (Mailboxer): storage for local messages

        """
        self.reps = reps if reps is not None else decking.Deck()
        self.cues = cues if cues is not None else decking.Deck()

        self.hby = hby
        self.aids = aids
        self.mbx = mbx if mbx is not None else Mailboxer(name=self.hby.name)
        self.postman = forwarding.Poster(hby=self.hby, mbx=self.mbx)

        doers = [self.postman, doing.doify(self.responseDo), doing.doify(self.cueDo)]
        super(Respondant, self).__init__(doers=doers, **kwa)

    def responseDo(self, tymth=None, tock=0.0):
        """
        Doifiable Doist compatibile generator method to process response messages from `exn` handlers.
        If dest is not in local environment, ignore the response (for now).  If dest has witnesses,
        pick one at random and send the response to that witness for storage in the recipients mailbox
        on that witness.  Otherwise this is a peer to peer HTTP message and should be stored in a mailbox
        locally for the recipient.

        Usage:
            add result of doify on this method to doers list
        """

        # enter context
        self.wind(tymth)
        self.tock = tock
        _ = (yield self.tock)

        while True:
            while self.reps:
                rep = self.reps.popleft()
                sender = rep["src"]
                recipient = rep["dest"]
                exn = rep["rep"]
                topic = rep["topic"]

                if recipient not in self.hby.kevers:
                    logger.error("unable to reply, dest {} not found".format(recipient))
                    continue

                senderHab = self.hby.habs[sender]
                if senderHab.mhab:
                    forwardHab = senderHab.mhab
                else:
                    forwardHab = senderHab

                # sign the exn to get the signature
                eattach = senderHab.endorse(exn, last=False, pipelined=False)
                del eattach[:exn.size]
                self.postman.send(recipient, topic=topic, serder=exn, hab=forwardHab, attachment=eattach)

                yield self.tock  # throttle just do one cue at a time

            yield self.tock

    def cueDo(self, tymth=None, tock=0.0):
        """
         Returns doifiable Doist compatibile generator method (doer dog) to process
            Kevery and Tevery cues deque

        Usage:
            add result of doify on this method to doers list
        """
        # enter context
        self.wind(tymth)
        self.tock = tock
        _ = (yield self.tock)

        while True:
            while not self.cues:
                yield self.tock

            cue = self.cues.pull() # self.cues.popleft()
            cueKin = cue["kin"]  # type or kind of cue
            if cueKin in ("receipt",):  # cue to receipt a received event from other pre
                serder = cue["serder"]  # Serder of received event for other pre
                cuedKed = serder.ked
                cuedPrefixer = coring.Prefixer(qb64=cuedKed["i"])

                # If respondant configured with list of acceptable AIDs to witness for, check them here
                if self.aids is not None and cuedPrefixer.qb64 not in self.aids:
                    continue

                if cuedPrefixer.qb64 in self.hby.kevers:
                    kever = self.hby.kevers[cuedPrefixer.qb64]
                    owits = oset(kever.wits)
                    if match := owits.intersection(self.hby.prefixes):
                        pre = match.pop()
                        hab = self.hby.habByPre(pre)
                        if hab is None:
                            continue

                        raw = hab.receipt(serder)
                        rserder = serdering.SerderKERI(raw=raw)
                        del raw[:rserder.size]
                        self.postman.send(serder.pre, topic="receipt", serder=rserder, hab=hab, attachment=raw)

            elif cueKin in ("replay",):
                src = cue["src"]
                dest = cue["dest"]
                msgs = cue["msgs"]

                hab = self.hby.habByPre(src)
                if hab is None:
                    continue

                if dest not in self.hby.kevers:
                    continue

                for msg in msgs:
                    raw = bytearray(msg)
                    serder = serdering.SerderKERI(raw=raw)
                    del raw[:serder.size]
                    self.postman.send(dest, topic="replay", serder=serder, hab=hab, attachment=raw)

            elif cueKin in ("reply",):
                src = cue["src"]
                serder = cue["serder"]

                dest = cue["dest"]

                if dest not in self.hby.kevers:
                    continue

                hab = self.hby.habByPre(src)
                if hab is None:
                    continue

                atc = hab.endorse(serder)
                del atc[:serder.size]
                self.postman.send(hab=hab, dest=dest, topic="reply", serder=serder, attachment=atc)

            else:
                self.cues.push(cue)

            yield self.tock
